from datetime import datetime
import numpy as np
import numpy.testing as nptest
import pandas as pd
import pytest
from scipy import stats

import pytesmo.metrics
from pytesmo.metrics import *
from pytesmo.metrics._fast_pairwise import (
    _moments_welford,
    _pearsonr_from_moments,
)
from pytesmo.metrics.pairwise import (
    has_ci,
    no_ci,
)
import pytesmo.metrics.deprecated as deprecated


@pytest.fixture
def testdata():
    # the seed avoids random failure
    np.random.seed(0)

    # generate random data that is correlated with r = 0.8
    cov = np.array([[1, 0.8], [0.8, 1]])
    X = np.linalg.cholesky(cov) @ np.random.randn(2, 1000)
    x, y = X[0, :], X[1, :]
    y = 1.1 * y + 0.5
    return x, y


@pytest.fixture
def arange_testdata():
    x = np.arange(10, dtype=float)
    y = np.arange(10, dtype=float) + 2
    return x, y


def test_analytical_ci_availability(testdata):
    for funcname in has_ci:
        func = getattr(pytesmo.metrics, funcname)
        assert has_analytical_ci(func)

    for funcname in no_ci:
        func = getattr(pytesmo.metrics, funcname)
        assert not has_analytical_ci(func)


def test_analytical_cis(testdata):
    x, y = testdata
    for funcname in has_ci:
        func = getattr(pytesmo.metrics, funcname)
        m, lb, ub = with_analytical_ci(func, x, y)
        m10, lb10, ub10 = with_analytical_ci(func, x, y, alpha=0.1)
        assert m == m10
        assert lb < m
        assert m < ub
        assert lb < lb10
        assert ub > ub10


@pytest.mark.parametrize("method", ["BCa", "basic", "percentile"])
def test_bootstrap_cis_simple(method):
    np.random.seed(313)
    # a test case with bias where we have analytical CIs to compare and
    # convergence should work if everything is implemented correctly
    x = np.random.randn(20000) + 5
    y = np.random.randn(20000)
    b, lb, ub = with_analytical_ci(pytesmo.metrics.bias, x, y, alpha=0.1)
    # x and y have variance 1, so x - y has variance 2.
    # Therefore, b ~ N(5, 2/n)
    rv = stats.norm(loc=5, scale=np.sqrt(2 / len(x)))
    # Approximate 95% intervals are therefore
    # 5 +- 2 * sqrt(2/n) = 5 +- 2 * sqrt(2/20000) ~ 5 +- 0.02
    assert abs(b - 5) < 0.02
    # The difference in CIs is based on experience values
    nptest.assert_almost_equal(b - lb, 5 - rv.ppf(0.05), 4)
    nptest.assert_almost_equal(ub - b, rv.ppf(0.95) - 5, 4)
    # bias CIs are symmetric
    assert b - lb == ub - b

    # test bootstrapped CIs
    bs_b, bs_lb, bs_ub = with_bootstrapped_ci(
        pytesmo.metrics.bias, x, y, alpha=0.1, method=method
    )
    assert bs_b == b  # this is the same function call
    # The difference in CIs is based on experience values, and not quite as
    # good as with analytical estimates
    nptest.assert_almost_equal(bs_lb, lb, 2)
    nptest.assert_almost_equal(bs_ub, rv.ppf(0.95), 2)


@pytest.mark.slow
@pytest.mark.parametrize("method", ["BCa", "basic", "percentile"])
def test_bootstrapped_cis(method, testdata):
    x, y = testdata
    for funcname in has_ci:
        func = getattr(pytesmo.metrics, funcname)
        m, lb, ub = with_analytical_ci(func, x, y, alpha=0.1)
        m_bs, lb_bs, ub_bs = with_bootstrapped_ci(
            func, x, y, alpha=0.1, nsamples=1000, method=method
        )
        assert m == m_bs
        assert lb_bs < ub_bs
        if funcname != "nrmsd":
            # nrmsd is a bit unstable when bootstrapping, due to the data
            # dependent normalization that is applied
            assert abs(ub - ub_bs) < 2e-2
            assert abs(lb - lb_bs) < 2e-2
        else:
            assert abs(ub - ub_bs) < 1e-1
            assert abs(lb - lb_bs) < 1e-1
    for funcname in no_ci:
        m, lb, ub = with_bootstrapped_ci(
            func, x, y, alpha=0.1, nsamples=1000, method=method
        )
        assert lb < m
        assert m < ub


# expected values of the metrics, to test whether it works
def test_expected_values(testdata):
    x, y = testdata

    expected_msd_bias = 0.5 ** 2
    expected_msd_var = 0.1 ** 2
    expected_msd_corr = 2 * 1 * 1.1 * (1 - 0.8)
    expected_msd = expected_msd_bias + expected_msd_corr + expected_msd_var
    expected_values = {
        "bias": -0.5,
        "msd": expected_msd,
        "rmsd": np.sqrt(expected_msd),  # actually not totally true
        "ubrmsd": np.sqrt(expected_msd_corr + expected_msd_var),  # not quite
        "mse_corr": expected_msd_corr,
        "mse_bias": expected_msd_bias,
        "mse_var": expected_msd_var,
        "pearson_r": 0.8,
    }

    for metric in expected_values:
        func = getattr(pytesmo.metrics, metric)
        m, lb, ub = with_bootstrapped_ci(func, x, y)
        e = expected_values[metric]
        assert lb < e
        assert e < ub


def test_tcol_metrics():
    n = 1000000

    mean_signal = 0.3
    sig_signal = 0.2
    signal = np.random.normal(mean_signal, sig_signal, n)

    sig_err_x = 0.02
    sig_err_y = 0.07
    sig_err_z = 0.04
    err_x = np.random.normal(0, sig_err_x, n)
    err_y = np.random.normal(0, sig_err_y, n)
    err_z = np.random.normal(0, sig_err_z, n)

    alpha_y = 0.2
    alpha_z = 0.5

    beta_y = 0.9
    beta_z = 1.6

    x = signal + err_x
    y = alpha_y + beta_y * (signal + err_y)
    z = alpha_z + beta_z * (signal + err_z)

    beta_pred = 1.0 / np.array((1, beta_y, beta_z))
    err_pred = np.array((sig_err_x, sig_err_y, sig_err_z))
    snr_pred = np.array(
        (
            (sig_signal / sig_err_x),
            (sig_signal / sig_err_y),
            (sig_signal / sig_err_z),
        )
    )

    snr, err, beta = tcol_metrics(x, y, z, ref_ind=0)

    nptest.assert_almost_equal(beta, beta_pred, decimal=2)
    nptest.assert_almost_equal(err, err_pred, decimal=2)
    nptest.assert_almost_equal(
        np.sqrt(10 ** (snr / 10.0)), snr_pred, decimal=1
    )


@pytest.mark.slow
def test_tcol_metrics_bootstrapped_cis():
    n = 10000

    mean_signal = 0.3
    sig_signal = 0.2
    signal = np.random.normal(mean_signal, sig_signal, n)

    sig_err_x = 0.02
    sig_err_y = 0.07
    sig_err_z = 0.04
    err_x = np.random.normal(0, sig_err_x, n)
    err_y = np.random.normal(0, sig_err_y, n)
    err_z = np.random.normal(0, sig_err_z, n)

    alpha_y = 0.2
    alpha_z = 0.5

    beta_y = 0.9
    beta_z = 1.6

    x = signal + err_x
    y = alpha_y + beta_y * (signal + err_y)
    z = alpha_z + beta_z * (signal + err_z)

    beta_pred = 1.0 / np.array((1, beta_y, beta_z))
    err_pred = np.array((sig_err_x, sig_err_y, sig_err_z))
    snr_pred = np.array(
        (
            (sig_signal / sig_err_x),
            (sig_signal / sig_err_y),
            (sig_signal / sig_err_z),
        )
    )

    (snr_result, err_result, beta_result) = tcol_metrics_with_bootstrapped_ci(
        x, y, z
    )
    snr = snr_result[0]
    err = err_result[0]
    beta = beta_result[0]

    nptest.assert_allclose(beta, beta_pred, rtol=1e-2, atol=1e-2)
    nptest.assert_allclose(err, err_pred, rtol=1e-2, atol=1e-2)
    nptest.assert_allclose(
        np.sqrt(10 ** (snr / 10.0)), snr_pred, rtol=5e-2, atol=1e-2
    )

    # 0: value, 1: lower, 2: upper
    assert np.all(snr_result[1] < snr_result[0])
    assert np.all(snr_result[0] < snr_result[2])
    assert np.all(err_result[1] < err_result[0])
    assert np.all(err_result[0] < err_result[2])
    # equality is okay because the reference dataset has beta == 1 always
    assert np.all(beta_result[1] <= beta_result[0])
    assert np.all(beta_result[0] <= beta_result[2])


def test_tcol_metrics_bootstrap_bad_data():
    n_datasets = 3
    n = 1000
    r = 0.8
    C = np.ones((n_datasets, n_datasets)) * r
    for i in range(n_datasets):
        C[i, i] = 1
    A = np.linalg.cholesky(C)
    X = (A @ np.random.randn(n_datasets, n)).T

    x, y, z = X[:, 0], X[:, 1], X[:, 2]
    (snr_result, err_result, beta_result) = tcol_metrics_with_bootstrapped_ci(
        x, y, z
    )
    # 0: value, 1: lower, 2: upper
    assert np.all(snr_result[1] < snr_result[0])
    assert np.all(snr_result[0] < snr_result[2])
    assert np.all(err_result[1] < err_result[0])
    assert np.all(err_result[0] < err_result[2])
    # equality is okay because the reference dataset has beta == 1 always
    assert np.all(beta_result[1] <= beta_result[0])
    assert np.all(beta_result[0] <= beta_result[2])


def test_bias(arange_testdata):
    """
    Test for bias
    """
    # example 1
    x, y = arange_testdata

    b_pred = -2
    b_obs = bias(x, y)

    nptest.assert_equal(b_obs, b_pred)

    # example 2
    x = np.arange(10) * 1.0
    y = np.arange(20, 30) * 1.0

    b_pred = 20.0
    b_obs = bias(y, x)

    nptest.assert_equal(b_obs, b_pred)


def test_aad(arange_testdata):
    """
    Test for average absolute deviation
    """
    # example 1
    x, y = arange_testdata
    dev_pred = 2.0
    dev_obs = aad(x, y)

    nptest.assert_equal(dev_obs, dev_pred)

    # example 2, with outlier
    y[-1] = 201.0
    dev_pred = 21.0
    dev_obs = aad(x, y)

    nptest.assert_equal(dev_obs, dev_pred)


def test_mad(arange_testdata):
    """
    Test for median absolute deviation
    """
    # example 1
    x, y = arange_testdata
    dev_pred = 2.0
    dev_obs = mad(x, y)

    nptest.assert_equal(dev_obs, dev_pred)

    # example 2, with outlier
    y[-1] = 201.0
    dev_pred = 2.0
    dev_obs = mad(x, y)

    nptest.assert_equal(dev_obs, dev_pred)


def test_rmsd(arange_testdata):
    """
    Test for rmsd
    """
    # example 1
    x, y = arange_testdata

    rmsd_pred = 2.0
    rmsd_obs = rmsd(x, y)

    nptest.assert_equal(rmsd_obs, rmsd_pred)

    # example 2, with outlier
    y[-1] = 100.0

    rmsd_pred = np.sqrt(831.7)
    rmsd_obs = rmsd(x, y)

    nptest.assert_almost_equal(rmsd_obs, rmsd_pred, 6)


def test_ubrmsd(arange_testdata):
    """
    Test for ubrmsd
    """
    # example 1
    x, y = arange_testdata

    ubrmsd_pred = 0
    ubrmsd_obs = ubrmsd(x, y)

    nptest.assert_equal(ubrmsd_obs, ubrmsd_pred)
    # aslo check consistency with direct formula
    ubrmsd_direct = np.sqrt(rmsd(x, y) ** 2 - bias(x, y) ** 2)
    nptest.assert_equal(ubrmsd_obs, ubrmsd_direct)

    # example 2, with outlier
    y[-1] = 100.0

    ubrmsd_pred = 26.7
    ubrmsd_obs = ubrmsd(x, y)

    nptest.assert_almost_equal(ubrmsd_obs, ubrmsd_pred, 6)
    # aslo check consistency with direct formula
    ubrmsd_direct = np.sqrt(rmsd(x, y) ** 2 - bias(x, y) ** 2)
    nptest.assert_almost_equal(ubrmsd_obs, ubrmsd_direct)


def test_msd(arange_testdata):
    """
    Test for msd
    """
    # example 1
    x, y = arange_testdata

    mse_pred = 4.0
    mse_bias_pred = 2.0 ** 2
    mse_obs = msd(x, y)
    mse_bias_obs = mse_bias(x, y)

    nptest.assert_equal(mse_obs, mse_pred)
    nptest.assert_equal(mse_bias_obs, mse_bias_pred)

    # example 2, with outlier
    y[-1] = 51.0

    mse_pred = 180.0
    mse_bias_pred = 36.0
    mse_obs = msd(x, y)
    mse_bias_obs = mse_bias(x, y)

    nptest.assert_almost_equal(mse_obs, mse_pred, 6)
    nptest.assert_almost_equal(mse_bias_obs, mse_bias_pred, 6)


def test_mse_decomposition(arange_testdata):
    # example 1
    x, y = arange_testdata

    mse_pred = 4.0
    mse_bias_pred = 2.0 ** 2
    mse_obs, _, mse_bias, _ = mse_decomposition(x, y)

    nptest.assert_equal(mse_obs, mse_pred)
    nptest.assert_equal(mse_bias, mse_bias_pred)

    # example 2, with outlier
    y[-1] = 51.0

    mse_pred = 180.0
    mse_bias_pred = 36.0
    mse_obs, _, mse_bias, _ = mse_decomposition(x, y)

    nptest.assert_almost_equal(mse_obs, mse_pred, 6)
    nptest.assert_almost_equal(mse_bias, mse_bias_pred, 6)


def test_rmsd_mse():
    """
    Test for rmsd and mse
    """
    # example 1
    x = np.random.randn(1000)
    y = np.random.randn(1000)

    rmsd_pred = rmsd(x, y)
    mse_pred, _, _, _ = mse_decomposition(x, y)

    nptest.assert_almost_equal(rmsd_pred ** 2, mse_pred, 6)


def test_pearsonr_from_moments():
    x = np.random.randn(100)
    y = np.random.randn(100) + x

    R_sp, p_R_sp = stats.pearsonr(x, y)

    covmatrix = np.cov(x, y)
    n = len(x)
    varx = covmatrix[0, 0]
    vary = covmatrix[1, 1]
    cov = covmatrix[1, 0]
    R, p_R = _pearsonr_from_moments(varx, vary, cov, n)

    nptest.assert_almost_equal(R_sp, R, 15)
    nptest.assert_almost_equal(p_R_sp, p_R, 15)


def test_rolling_pr_rmsd():
    # setup test data
    window_size = "30d"
    min_periods = 2
    center = True

    startdate = datetime(2000, 1, 1)
    enddate = datetime(2000, 12, 31)
    dt_index = pd.date_range(start=startdate, end=enddate, freq="D")

    names = ["ref", "k1", "k2", "k3"]
    # always 0.5
    np.random.seed(10)
    df = pd.DataFrame(
        index=dt_index,
        data={
            name: 0.5 + 0.01 * np.random.randn(dt_index.size) for name in names
        },
    )

    df["k1"] += 0.2  # some positive bias
    df["k2"] -= 0.2  # some negative bias
    df["k3"] -= 0.3  # some more negative bias

    data = df
    xy = data.to_numpy()
    timestamps = data.index.to_julian_date().values
    window_size_jd = pd.Timedelta(window_size).to_numpy() / np.timedelta64(
        1, "D"
    )

    pr, rmsd = deprecated.rolling_pr_rmsd(
        timestamps, xy, window_size_jd, center, min_periods
    )

    prnew, rmsdnew = rolling_pr_rmsd(
        timestamps, xy[:, 0], xy[:, 1], window_size_jd, center, min_periods
    )

    nptest.assert_almost_equal(pr[:, 0], prnew[:, 0])
    # p-values might not match up to machine precision, especially if they are
    # close to 1
    nptest.assert_almost_equal(pr[:, 1], prnew[:, 1], 4)
    nptest.assert_almost_equal(rmsd, rmsdnew)


def test_moments_welford():
    x = np.random.randn(1000)
    y = np.random.randn(1000)

    mx, my, varx, vary, cov = _moments_welford(x, y)

    nptest.assert_almost_equal(np.mean(x), mx, 14)
    nptest.assert_almost_equal(np.mean(y), my, 14)
    nptest.assert_almost_equal(np.var(x), varx, 14)
    nptest.assert_almost_equal(np.var(y), vary, 14)
    nptest.assert_almost_equal(np.cov(x, y, ddof=0)[0, 1], cov, 14)



try:
    from IPython import parallel
except ImportError:
    pass
try:
    from itertools import izip as zip
except ImportError:
    # python 3
    pass
from datetime import datetime

from pytesmo.validation_framework.results_manager import netcdf_results_manager


def func(job):
    """
    Function which calls the start_processing method implemented in setup_code.
    """
    return start_processing(job)


def start_validation(setup_code):
    """
    Perform the validation with IPython parallel processing.

    Parameters
    ----------
    setup_code : string
        Path to .py file containing the setup for the validation.
    """
    c = parallel.Client()
    dv = c[:]
    lview = c.load_balanced_view()

    dv.run(setup_code, block=True)

    jobs = None
    try:
        jobs = dv['jobs'][0]
    except parallel.CompositeError:
        print("Variable 'jobs' is not defined!")

    save_path = None
    try:
        save_path = dv['save_path'][0]
    except parallel.CompositeError:
        print("Variable 'save_path' is not defined!")

    to_write = len(jobs)
    if (jobs is not None) and (save_path is not None):
        with lview.temp_flags(retries=2):
            amr = lview.map_async(func, jobs)
            results = zip(amr, jobs)
            for result, job in results:
                netcdf_results_manager(result, save_path)
                to_write -= 1
                print('job = ' + str(job), 'remaining jobs = ' + str(to_write))

    c[:].clear()


if __name__ == '__main__':

    start = datetime.now()
    print('Start Validation')

    # Note that before starting the validation you must start a controller
    # and engines, for example by using: ipcluster start -n 4
    # This command will launch a controller and 4 engines on the local machine.
    # Also, do not forget to change the setup_code path to your current setup.
    setup_code = "/media/sf_H/swdvlp/github/pytesmo/examples/setup_validation_ASCAT_ISMN.py"
    start_validation(setup_code)

    print('Elapsed time:', datetime.now() - start)



'''
Created on June 20, 2013
'''
import warnings

import pandas as pd
import numpy as np
from cadati.conv_doy import doy
from cadati.jd_date import julian2date
from pytesmo.time_series.filtering import moving_average


def calc_anomaly(Ser,
                 window_size=35,
                 climatology=None,
                 respect_leap_years=True,
                 return_clim=False):
    '''
    Calculates the anomaly of a time series (Pandas series).
    Both, climatology based, or moving-average based anomalies can be
    calculated

    Parameters
    ----------
    Ser : pandas.Series (index must be a DateTimeIndex)

    window_size : float, optional
        The window-size [days] of the moving-average window to calculate the
        anomaly reference (only used if climatology is not provided)
        Default: 35 (days)

    climatology : pandas.Series (index: 1-366), optional
        if provided, anomalies will be based on the climatology

    timespan : [timespan_from, timespan_to], datetime.datetime(y,m,d), optional
        If set, only a subset

    respect_leap_years : boolean, optional
        If set then leap years will be respected during matching of the climatology
        to the time series

    return_clim : boolean, optional
        if set to true the return argument will be a DataFrame which
        also contains the climatology time series.
        Only has an effect if climatology is used.

    Returns
    -------
    anomaly : pandas.Series
        Series containing the calculated anomalies
    '''

    if climatology is not None:

        if type(Ser.index) == pd.DatetimeIndex:

            year, month, day = (np.asarray(Ser.index.year),
                                np.asarray(Ser.index.month),
                                np.asarray(Ser.index.day))

        else:
            year, month, day = julian2date(Ser.index.values)[0:3]

        if respect_leap_years:
            doys = doy(month, day, year)
        else:
            doys = doy(month, day)

        df = pd.DataFrame()
        df['absolute'] = Ser
        df['doy'] = doys

        clim = pd.DataFrame({'climatology': climatology})

        df = df.join(clim, on='doy', how='left')

        anomaly = df['absolute'] - df['climatology']
        anomaly.index = df.index

        if return_clim:
            anomaly = pd.DataFrame({'anomaly': anomaly})
            anomaly['climatology'] = df['climatology']

    else:
        reference = moving_average(Ser, window_size=window_size)
        anomaly = Ser - reference

    return anomaly


def _index_units(year,
                 month,
                 day,
                 unit="day",
                 respect_leap_years=True) -> (np.array, int):
    """Provide the indices for the specified unit type and the index lenth"""
    if respect_leap_years:
        args = month, day, year
    else:
        args = month, day

    if unit == "day":
        return doy(*args), 366
    elif unit == "month":
        return month, 12
    else:
        raise ValueError(f"Invalid unit: {unit}")


def calc_climatology(Ser,
                     moving_avg_orig=5,
                     moving_avg_clim=None,
                     median=False,
                     timespan=None,
                     fill=np.nan,
                     wraparound=True,
                     respect_leap_years=False,
                     interpolate_leapday=False,
                     fillna=True,
                     min_obs_orig=1,
                     min_obs_clim=1,
                     output_freq="day"):
    """
    Calculates the climatology of a data set.

    Parameters
    ----------
    Ser : pandas.Series (index must be a DateTimeIndex or julian date)

    moving_avg_orig : float, optional
        The size of the moving_average window [days] that will be applied on the
        input Series (gap filling, short-term rainfall correction)
        Default: 5

    moving_avg_clim : float, optional
        The size of the moving_average window in days that will be applied on the
        calculated climatology (long-term event correction). If None is passed, it
        will be calculated from the 'output_freq' value:
            - 'day': 35
            - 'month': 3
        Default: None

    median : boolean, optional
        if set to True, the climatology will be based on the median conditions

    timespan : [timespan_from, timespan_to], datetime.datetime(y,m,d), optional
        Set this to calculate the climatology based on a subset of the input
        Series

    fill : float or int, optional
        Fill value to use for days on which no climatology exists

    wraparound : boolean, optional
        If set then the climatology is wrapped around at the edges before
        doing the second running average (long-term event correction)

    respect_leap_years : boolean, optional
        If set then leap years will be respected during the calculation of
        the climatology. Only valid with 'unit' value set to 'day'.
        Default: False

    interpolate_leapday: boolean, optional
        <description>. Only valid with 'unit' value set to 'day'.
        Default: False

    fillna: boolean, optional
        If set, then the moving average used for the calculation of the
        climatology will be filled at the nan-values

    min_obs_orig: int
        Minimum observations required to give a valid output in the first
        moving average applied on the input series

    min_obs_clim: int
        Minimum observations required to give a valid output in the second
        moving average applied on the calculated climatology

    output_freq: str, optional
        Determines the output frequency (time unit) of the climatology
        calculation (independently of the 'Ser' input frequency).
        Currently, supported options are 'day', 'month'.
        Default: 'day'

    Returns
    -------
    climatology : pandas.Series
        Containing the calculated climatology. The size of the series depends on
        the type of climatology being calculated, based on the value of 'output_freq':
            - 366 values for a daily climatology, behaving as a leap year
            - 12 values for a monthly climatology
    """
    # establish the moving window size
    default_moving_avg_clim = {"day": 35, "month": 3}

    if moving_avg_clim is None:
        moving_avg_clim = default_moving_avg_clim[output_freq]

    if output_freq == "month" and moving_avg_clim > 5:
        # in case someone changes unit but not moving_avg_clim a warning might be useful
        warnings.warn(
            f"Window for moving average of climatology set to {moving_avg_clim} months, is this intentional?"
        )

    if output_freq != "day":
        # irrelevant at lower frequencies than daily
        respect_leap_years, interpolate_leapday = False, False

    if timespan is not None:
        Ser = Ser.truncate(before=timespan[0], after=timespan[1])

    Ser = moving_average(
        Ser, window_size=moving_avg_orig, fillna=fillna, min_obs=min_obs_orig)

    Ser = pd.DataFrame(Ser)

    if type(Ser.index) == pd.DatetimeIndex:
        year, month, day = (np.asarray(Ser.index.year),
                            np.asarray(Ser.index.month),
                            np.asarray(Ser.index.day))
    else:
        year, month, day = julian2date(Ser.index.values)[0:3]

    # provide indices for the selected unit
    indices, n_idx = _index_units(
        year,
        month,
        day,
        unit=output_freq,
        respect_leap_years=respect_leap_years)
    Ser['unit'] = indices

    if median:
        clim = Ser.groupby('unit').median()
    else:
        clim = Ser.groupby('unit').mean()

    clim_ser = pd.Series(clim.values.flatten(), index=clim.index.values)

    clim_ser = clim_ser.reindex(np.arange(n_idx) + 1)

    if wraparound:
        index_old = clim_ser.index.copy()
        left_mirror = clim_ser.iloc[-moving_avg_clim:]
        right_mirror = clim_ser.iloc[:moving_avg_clim]
        # Shift index to start at n_idx - index at -moving_avg_clim
        # to run over a whole year while keeping gaps the same size
        right_mirror.index = right_mirror.index + n_idx * 2
        clim_ser.index = clim_ser.index + n_idx
        clim_ser = pd.concat([left_mirror, clim_ser, right_mirror])

        clim_ser = moving_average(
            clim_ser,
            window_size=moving_avg_clim,
            fillna=fillna,
            min_obs=min_obs_clim)
        clim_ser = clim_ser.iloc[moving_avg_clim:-moving_avg_clim]
        clim_ser.index = index_old
    else:
        clim_ser = moving_average(
            clim_ser,
            window_size=moving_avg_clim,
            fillna=fillna,
            min_obs=min_obs_clim)

    # keep hardcoding as it's only for doys
    if interpolate_leapday and not respect_leap_years:
        clim_ser[60] = np.mean((clim_ser[59], clim_ser[61]))
    elif interpolate_leapday and respect_leap_years:
        clim_ser[366] = np.mean((clim_ser[365], clim_ser[1]))

    clim_ser = clim_ser.fillna(fill)

    return clim_ser



# coding: utf-8
# Copyright (c) 2015,Vienna University of Technology,
# Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of the Vienna University of Technology,
#      Department of Geodesy and Geoinformation nor the
#      names of its contributors may be used to endorse or promote products
#      derived from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
Tests for the validation framework
Created on Mon Jul  6 12:49:07 2015
"""

from datetime import datetime
import netCDF4 as nc
import numpy as np
import numpy.testing as nptest
import os
import pytest
import tempfile
import warnings
import pandas as pd
import pytest

import pygeogrids.grids as grids
from pygeobase.io_base import GriddedTsBase

import pytesmo.validation_framework.temporal_matchers as temporal_matchers
import pytesmo.validation_framework.metric_calculators as metrics_calculators
from pytesmo.validation_framework.results_manager import netcdf_results_manager
from pytesmo.validation_framework.data_manager import DataManager
from pytesmo.validation_framework.results_manager import PointDataResults
from pytesmo.validation_framework.validation import Validation
from pytesmo.validation_framework.validation import args_to_iterable
import pytesmo.validation_framework.error_handling as eh

from pytesmo.validation_framework.metric_calculators import (
    PairwiseIntercomparisonMetrics)
from pytesmo.validation_framework.temporal_matchers import (
    make_combined_temporal_matcher)

from ismn.interface import ISMN_Interface

with warnings.catch_warnings():
    warnings.filterwarnings("ignore")
    from ascat.read_native.cdr import AscatGriddedNcTs

from .utils import create_datasets

if __name__ != "__main__":
    from tests.test_validation_framework.test_datasets import (
        setup_TestDatasets,
        setup_two_without_overlap,
        setup_three_with_two_overlapping,
        MaskingTestDataset,
    )


@pytest.fixture
def ascat_reader():
    ascat_data_folder = os.path.join(
        os.path.dirname(__file__),
        "..",
        "test-data",
        "sat",
        "ascat",
        "netcdf",
        "55R22",
    )

    ascat_grid_folder = os.path.join(
        os.path.dirname(__file__),
        "..",
        "test-data",
        "sat",
        "ascat",
        "netcdf",
        "grid",
    )
    grid_fname = os.path.join(ascat_grid_folder, "TUW_WARP5_grid_info_2_1.nc")

    static_layers_folder = os.path.join(
        os.path.dirname(__file__),
        "..",
        "test-data",
        "sat",
        "h_saf",
        "static_layer",
    )

    ascat_reader = AscatGriddedNcTs(
        ascat_data_folder,
        "TUW_METOP_ASCAT_WARP55R22_{:04d}",
        grid_filename=grid_fname,
        static_layer_path=static_layers_folder,
        ioclass_kws={'read_bulk': True}
    )
    return ascat_reader


@pytest.fixture
def ismn_reader():
    # Initialize ISMN reader

    ismn_data_folder = os.path.join(
        os.path.dirname(__file__),
        "..",
        "test-data",
        "ismn",
        "multinetwork",
        "header_values",
    )
    meta_path = tempfile.mkdtemp()
    ismn_reader = ISMN_Interface(ismn_data_folder, meta_path=meta_path)

    return ismn_reader


def check_results(
    filename: str,
    target_vars: dict,
    variables: list = None,
):
    """
    Check that standard vars are present and that nobs, rho and rmsd match
    the given values. Vars can be optionally specified
    """
    if variables is not None:
        vars_should = variables
    else:
        vars_should = [
            u"n_obs", u"tau", u"gpi", u"RMSD", u"lon", u"p_tau", u"BIAS",
            u"p_rho", u"rho", u"lat", u"R", u"p_R", u"time", u"idx",
            u"_row_size", u"status"
        ]

    with nc.Dataset(filename, mode="r") as results:
        vars = results.variables.keys()
        assert sorted(vars) == sorted(vars_should)

        for varname, should_values in target_vars.items():
            values = results.variables[varname][:]
            if varname == "n_obs":
                values = list(values)
                assert sorted(values) == sorted(should_values)
            elif varname == "network":
                nptest.assert_equal(sorted(values), sorted(should_values))
            else:
                nptest.assert_allclose(
                    sorted(values), sorted(should_values), rtol=1e-4)


@pytest.mark.slow
@pytest.mark.full_framework
def test_ascat_ismn_validation(ascat_reader, ismn_reader):
    """
    Test processing framework with some ISMN and ASCAT sample data
    """
    jobs = []

    ids = ismn_reader.get_dataset_ids(
        variable="soil_moisture", min_depth=0, max_depth=0.1)
    for idx in ids:
        metadata = ismn_reader.read_metadata(idx)
        jobs.append((idx, metadata["longitude"].val, metadata["latitude"].val))


    # Create the validation object.

    datasets = {
        "ISMN": {
            "class": ismn_reader,
            "columns": ["soil_moisture"]
        },
        "ASCAT": {
            "class": ascat_reader,
            "columns": ["sm"],
            "kwargs": {
                "mask_frozen_prob": 80,
                "mask_snow_prob": 80,
                "mask_ssf": True,
            },
        },
    }

    read_ts_names = {"ASCAT": "read", "ISMN": "read"}
    period = [datetime(2007, 1, 1), datetime(2014, 12, 31)]

    datasets = DataManager(
        datasets, "ISMN", period, read_ts_names=read_ts_names)

    process = Validation(
        datasets,
        "ISMN",
        temporal_ref="ASCAT",
        scaling="cdf_match",
        scaling_ref="ASCAT",
        metrics_calculators={
            (2, 2):
                metrics_calculators.BasicMetrics(other_name="k1").calc_metrics
        },
        period=period,
    )

    # Create the variable ***save_path*** which is a string representing the
    # path where the results will be saved. **DO NOT CHANGE** the name
    # ***save_path*** because it will be searched during the parallel
    # processing!
    #

    with tempfile.TemporaryDirectory() as save_path:
        for job in jobs:
            results = process.calc(*job)
            netcdf_results_manager(results, save_path)

        results_fname = os.path.join(save_path,
                                     "ASCAT.sm_with_ISMN.soil_moisture.nc")
        # targets
        target_vars = {
            "n_obs": [357, 384, 1646, 1875, 1915, 467, 141, 251],
            "rho":
                np.array([
                    0.5393457276100486,
                    0.7002289363385688,
                    0.6220023416557768,
                    0.5364715553916934,
                    0.3041366641200411,
                    0.6759989589212635,
                    0.8418980829611084,
                    0.7391369789045885,
                ], dtype=np.float32),
            "RMSD":
                np.array([
                    11.577185,
                    7.788051,
                    17.73936,
                    21.145535,
                    14.092733,
                    13.741866,
                    12.1061,
                    13.058796,
                ], dtype=np.float32)
        }

        check_results(
            filename=results_fname,
            target_vars=target_vars,
        )

        ascat_reader.close()
        del ascat_reader


@pytest.mark.slow
@pytest.mark.full_framework
def test_ascat_ismn_validation_metadata(ascat_reader, ismn_reader):
    """
    Test processing framework with some ISMN and ASCAT sample data
    """
    jobs = []

    ids = ismn_reader.get_dataset_ids(
        variable="soil_moisture", min_depth=0, max_depth=0.1)

    metadata_dict_template = {
        "network": np.array(["None"], dtype="U256"),
        "station": np.array(["None"], dtype="U256"),
        "landcover": np.float32([np.nan]),
        "climate": np.array(["None"], dtype="U4"),
    }

    for idx in ids:
        metadata = ismn_reader.read_metadata(idx, 'obj')
        metadata_dict = [{
            "network": metadata["network"][1],
            "station": metadata["station"][1],
            "landcover": metadata["lc_2010"][1],
            "climate": metadata["climate_KG"][1],
        }]
        jobs.append((idx, metadata["longitude"][1], metadata["latitude"][1],
                     metadata_dict))

    # Create the variable ***save_path*** which is a string representing the
    # path where the results will be saved. **DO NOT CHANGE** the name
    # ***save_path*** because it will be searched during the parallel
    # processing!

    save_path = tempfile.mkdtemp()

    # Create the validation object.

    datasets = {
        "ISMN": {
            "class": ismn_reader,
            "columns": ["soil_moisture"],
        },
        "ASCAT": {
            "class": ascat_reader,
            "columns": ["sm"],
            "kwargs": {
                "mask_frozen_prob": 80,
                "mask_snow_prob": 80,
                "mask_ssf": True,
            },
        },
    }

    read_ts_names = {"ASCAT": "read", "ISMN": "read"}
    period = [datetime(2007, 1, 1), datetime(2014, 12, 31)]

    datasets = DataManager(
        datasets, "ISMN", period, read_ts_names=read_ts_names)
    process = Validation(
        datasets,
        "ISMN",
        temporal_ref="ASCAT",
        scaling="cdf_match",
        scaling_ref="ASCAT",
        metrics_calculators={
            (2, 2):
                metrics_calculators.BasicMetrics(
                    other_name="k1",
                    metadata_template=metadata_dict_template).calc_metrics
        },
        period=period,
    )

    for job in jobs:
        results = process.calc(*job)
        netcdf_results_manager(results, save_path)

    results_fname = os.path.join(save_path,
                                 "ASCAT.sm_with_ISMN.soil_moisture.nc")
    target_vars = {
        "n_obs": [
            357,
            384,
            1646,
            1875,
            1915,
            467,
            141,
            251
        ],
        # reference values for rho were inferred from the unscaled rho, since
        # the CDF scaling should not change rank correlation
        "rho": np.array([
            0.5393457276100486,
            0.7002289363385688,
            0.6220023416557768,
            0.5364715553916934,
            0.3041366641200411,
            0.6759989589212635,
            0.8418980829611084,
            0.7391369789045885,
        ], dtype=np.float32),
        # reference values for RMSD were taken from the first run with the new
        # CDF matching implementation
        "RMSD": np.array([
            11.577185,
            7.788051,
            17.73936,
            21.145535,
            14.092733,
            13.741866,
            12.1061,
            13.058796,
        ], dtype=np.float32),
        "network": np.array([
            "MAQU",
            "MAQU",
            "SCAN",
            "SCAN",
            "SCAN",
            "SOILSCAPE",
            "SOILSCAPE",
            "SOILSCAPE",
        ], dtype="U256",)
    }
    vars_should = [
        'BIAS', 'R', 'RMSD', '_row_size', 'climate', 'gpi', 'idx', 'landcover',
        'lat', 'lon', 'n_obs', 'network', 'p_R', 'p_rho', 'p_tau', 'rho',
        'station', 'tau', 'time', 'status'
    ]

    check_results(
        filename=results_fname, target_vars=target_vars, variables=vars_should)

    ascat_reader.close()


@pytest.mark.slow
@pytest.mark.full_framework
def test_validation_with_averager(ascat_reader, ismn_reader):
    """
    Test processing framework with averaging module. ASCAT and ISMN data are
    used here with no geographical
    considerations (the lut is provided more upstream and contains this
    information already)
    """
    while hasattr(ascat_reader, 'cls'):
        ascat_reader = ascat_reader.cls
    # lookup table between the ascat and ismn points
    # - not geographically correct
    upscaling_lut = {
        "ISMN": {
            1814367: [(0, 102.1333, 33.8833), (1, 102.1333, 33.6666)],
            1803695: [(2, -86.55, 34.783), (3, -97.083, 37.133),
                      (4, -105.417, 34.25)],
            1856312: [(5, -120.9675, 38.43003), (6, -120.78559, 38.14956),
                      (7, -120.80639, 38.17353)]
        }
    }
    gpis = (1814367, 1803695, 1856312)
    lons, lats = [], []
    for gpi in gpis:
        lon, lat = ascat_reader.grid.gpi2lonlat(gpi)
        lons.append(lon)
        lats.append(lat)

    jobs = [(gpis, lons, lats)]

    # Create the variable ***save_path*** which is a string representing the
    # path where the results will be saved. **DO NOT CHANGE** the name
    # ***save_path*** because it will be searched during the parallel
    # processing!

    save_path = tempfile.mkdtemp()

    # Create the validation object.

    datasets = {
        "ASCAT": {
            "class": ascat_reader,
            "columns": ["sm"],
            "kwargs": {
                "mask_frozen_prob": 80,
                "mask_snow_prob": 80,
                "mask_ssf": True,
            }
        },
        "ISMN": {
            "class": ismn_reader,
            "columns": ["soil_moisture"],
        },
    }

    read_ts_names = {"ASCAT": "read", "ISMN": "read"}
    period = [datetime(2007, 1, 1), datetime(2014, 12, 31)]

    datasets = DataManager(
        datasets,
        "ASCAT",
        period,
        read_ts_names=read_ts_names,
        upscale_parms={
            "upscaling_method": "average",
            "temporal_stability": True,
            "upscaling_lut": upscaling_lut,
        },
    )
    process = Validation(
        datasets,
        "ASCAT",
        temporal_ref="ISMN",
        scaling="cdf_match",
        scaling_ref="ISMN",
        metrics_calculators={
            (2, 2):
                metrics_calculators.BasicMetrics(other_name="k1").calc_metrics
        },
        period=period,
    )

    for job in jobs:
        results = process.calc(*job)
        netcdf_results_manager(results, save_path)

    results_fname = os.path.join(save_path,
                                 "ASCAT.sm_with_ISMN.soil_moisture.nc")

    target_vars = {
        "n_obs": [
            764,
            2392,
            904
        ],
        "rho": np.array(
            [-0.012487,
             0.255156,
             0.635517,
             ], dtype=np.float32),
        "RMSD": np.array([
            0.05708659,
            0.05607013,
            0.11864747
        ], dtype=np.float32),
        "R": np.array([
            0.64745444,
            -0.0121021,
            0.2571774
        ], dtype=np.float32)
    }

    check_results(
        filename=results_fname,
        target_vars=target_vars,
    )

    ascat_reader.close()


def test_validation_error_n2_k2():
    datasets = setup_TestDatasets()

    dm = DataManager(
        datasets,
        "DS1",
        read_ts_names={d: "read" for d in ["DS1", "DS2", "DS3"]},
    )

    # n less than number of datasets is no longer allowed
    with pytest.raises(ValueError):
        Validation(
            dm,
            "DS1",
            temporal_matcher=temporal_matchers.BasicTemporalMatching(
                window=1 / 24.0).combinatory_matcher,
            scaling="cdf_match",
            metrics_calculators={
                (2, 2):
                    metrics_calculators.BasicMetrics(
                        other_name="k1").calc_metrics
            },
        )


def test_validation_n3_k2_temporal_matching_no_matches():
    tst_results = {
        (("DS1", "x"), ("DS2", "y")): {
            'gpi': np.array([4], dtype=np.int32),
            'lon': np.array([4.]),
            'lat': np.array([4.]),
            'status': np.array([eh.NO_TEMP_MATCHED_DATA], dtype=np.int32),
            'n_obs': np.array([0], dtype=np.int32),
            'R': np.array([np.nan], dtype=np.float32),
            'p_R': np.array([np.nan], dtype=np.float32),
            'rho': np.array([np.nan], dtype=np.float32),
            'p_rho': np.array([np.nan], dtype=np.float32),
            'RMSD': np.array([np.nan], dtype=np.float32),
            'BIAS': np.array([np.nan], dtype=np.float32),
            'tau': np.array([np.nan], dtype=np.float32),
            'p_tau': np.array([np.nan], dtype=np.float32)
        }
    }

    datasets = setup_two_without_overlap()

    dm = DataManager(
        datasets,
        "DS1",
        read_ts_names={d: "read" for d in ["DS1", "DS2", "DS3"]},
    )

    process = Validation(
        dm,
        "DS1",
        temporal_matcher=temporal_matchers.BasicTemporalMatching(
            window=1 / 24.0).combinatory_matcher,
        scaling="cdf_match",
        metrics_calculators={
            (3, 2):
                metrics_calculators.BasicMetrics(other_name="k1").calc_metrics
        },
    )

    jobs = process.get_processing_jobs()
    for job in jobs:
        with pytest.raises(eh.NoTempMatchedDataError):
            results = process.calc(*job, handle_errors="raise")

    jobs = process.get_processing_jobs()
    for job in jobs:
        results = process.calc(*job, handle_errors="ignore")
        assert sorted(list(results)) == sorted(list(tst_results))


def test_validation_n3_k2_data_manager_argument():
    tst_results = {
        (("DS1", "x"), ("DS3", "y")): {
            "n_obs": np.array([1000], dtype=np.int32),
            "tau": np.array([np.nan], dtype=np.float32),
            "gpi": np.array([4], dtype=np.int32),
            "RMSD": np.array([0.0], dtype=np.float32),
            "lon": np.array([4.0]),
            "p_tau": np.array([np.nan], dtype=np.float32),
            "BIAS": np.array([0.0], dtype=np.float32),
            "p_rho": np.array([0.0], dtype=np.float32),
            "rho": np.array([1.0], dtype=np.float32),
            "lat": np.array([4.0]),
            "R": np.array([1.0], dtype=np.float32),
            "p_R": np.array([0.0], dtype=np.float32),
        },
        (("DS1", "x"), ("DS2", "y")): {
            "n_obs": np.array([1000], dtype=np.int32),
            "tau": np.array([np.nan], dtype=np.float32),
            "gpi": np.array([4], dtype=np.int32),
            "RMSD": np.array([0.0], dtype=np.float32),
            "lon": np.array([4.0]),
            "p_tau": np.array([np.nan], dtype=np.float32),
            "BIAS": np.array([0.0], dtype=np.float32),
            "p_rho": np.array([0.0], dtype=np.float32),
            "rho": np.array([1.0], dtype=np.float32),
            "lat": np.array([4.0]),
            "R": np.array([1.0], dtype=np.float32),
            "p_R": np.array([0.0], dtype=np.float32),
        },
        (("DS1", "x"), ("DS3", "x")): {
            "n_obs": np.array([1000], dtype=np.int32),
            "tau": np.array([np.nan], dtype=np.float32),
            "gpi": np.array([4], dtype=np.int32),
            "RMSD": np.array([0.0], dtype=np.float32),
            "lon": np.array([4.0]),
            "p_tau": np.array([np.nan], dtype=np.float32),
            "BIAS": np.array([0.0], dtype=np.float32),
            "p_rho": np.array([0.0], dtype=np.float32),
            "rho": np.array([1.0], dtype=np.float32),
            "lat": np.array([4.0]),
            "R": np.array([1.0], dtype=np.float32),
            "p_R": np.array([0.0], dtype=np.float32),
        },
        (("DS2", "y"), ("DS3", "x")): {
            "gpi": np.array([4], dtype=np.int32),
            "lon": np.array([4.0]),
            "lat": np.array([4.0]),
            "n_obs": np.array([1000], dtype=np.int32),
            "R": np.array([1.0], dtype=np.float32),
            "p_R": np.array([0.0], dtype=np.float32),
            "rho": np.array([1.0], dtype=np.float32),
            "p_rho": np.array([0.0], dtype=np.float32),
            "RMSD": np.array([0.0], dtype=np.float32),
            "BIAS": np.array([0.0], dtype=np.float32),
            "tau": np.array([np.nan], dtype=np.float32),
            "p_tau": np.array([np.nan], dtype=np.float32),
        },
        (("DS2", "y"), ("DS3", "y")): {
            "gpi": np.array([4], dtype=np.int32),
            "lon": np.array([4.0]),
            "lat": np.array([4.0]),
            "n_obs": np.array([1000], dtype=np.int32),
            "R": np.array([1.0], dtype=np.float32),
            "p_R": np.array([0.0], dtype=np.float32),
            "rho": np.array([1.0], dtype=np.float32),
            "p_rho": np.array([0.0], dtype=np.float32),
            "RMSD": np.array([0.0], dtype=np.float32),
            "BIAS": np.array([0.0], dtype=np.float32),
            "tau": np.array([np.nan], dtype=np.float32),
            "p_tau": np.array([np.nan], dtype=np.float32),
        },
    }

    datasets = setup_TestDatasets()
    dm = DataManager(
        datasets,
        "DS1",
        read_ts_names={d: "read" for d in ["DS1", "DS2", "DS3"]},
    )

    process = Validation(
        dm,
        "DS1",
        temporal_matcher=temporal_matchers.BasicTemporalMatching(
            window=1 / 24.0).combinatory_matcher,
        scaling="cdf_match",
        metrics_calculators={
            (3, 2):
                metrics_calculators.BasicMetrics(other_name="k1").calc_metrics
        },
    )

    jobs = process.get_processing_jobs()
    for job in jobs:
        results = process.calc(*job)
        assert sorted(list(results)) == sorted(list(tst_results))

    datasets = setup_TestDatasets()
    dm = DataManager(
        datasets,
        "DS1",
        read_ts_names={d: "read" for d in ["DS1", "DS2", "DS3"]},
    )

    process = Validation(
        dm,
        "DS1",
        temporal_matcher=temporal_matchers.BasicTemporalMatching(
            window=1 / 24.0).combinatory_matcher,
        scaling="cdf_match",
        metrics_calculators={
            (3, 2):
                metrics_calculators.BasicMetrics(other_name="k1").calc_metrics
        },
    )

    jobs = process.get_processing_jobs()
    for job in jobs:
        results = process.calc(*job)
        assert sorted(list(results)) == sorted(list(tst_results))


def test_validation_n3_k2():
    tst_results = {
        (("DS1", "x"), ("DS3", "y")): {
            "n_obs": np.array([1000], dtype=np.int32),
            "tau": np.array([np.nan], dtype=np.float32),
            "gpi": np.array([4], dtype=np.int32),
            "RMSD": np.array([0.0], dtype=np.float32),
            "lon": np.array([4.0]),
            "p_tau": np.array([np.nan], dtype=np.float32),
            "BIAS": np.array([0.0], dtype=np.float32),
            "p_rho": np.array([0.0], dtype=np.float32),
            "rho": np.array([1.0], dtype=np.float32),
            "lat": np.array([4.0]),
            "R": np.array([1.0], dtype=np.float32),
            "p_R": np.array([0.0], dtype=np.float32),
        },
        (("DS1", "x"), ("DS2", "y")): {
            "n_obs": np.array([1000], dtype=np.int32),
            "tau": np.array([np.nan], dtype=np.float32),
            "gpi": np.array([4], dtype=np.int32),
            "RMSD": np.array([0.0], dtype=np.float32),
            "lon": np.array([4.0]),
            "p_tau": np.array([np.nan], dtype=np.float32),
            "BIAS": np.array([0.0], dtype=np.float32),
            "p_rho": np.array([0.0], dtype=np.float32),
            "rho": np.array([1.0], dtype=np.float32),
            "lat": np.array([4.0]),
            "R": np.array([1.0], dtype=np.float32),
            "p_R": np.array([0.0], dtype=np.float32),
        },
        (("DS1", "x"), ("DS3", "x")): {
            "n_obs": np.array([1000], dtype=np.int32),
            "tau": np.array([np.nan], dtype=np.float32),
            "gpi": np.array([4], dtype=np.int32),
            "RMSD": np.array([0.0], dtype=np.float32),
            "lon": np.array([4.0]),
            "p_tau": np.array([np.nan], dtype=np.float32),
            "BIAS": np.array([0.0], dtype=np.float32),
            "p_rho": np.array([0.0], dtype=np.float32),
            "rho": np.array([1.0], dtype=np.float32),
            "lat": np.array([4.0]),
            "R": np.array([1.0], dtype=np.float32),
            "p_R": np.array([0.0], dtype=np.float32),
        },
        (("DS2", "y"), ("DS3", "x")): {
            "gpi": np.array([4], dtype=np.int32),
            "lon": np.array([4.0]),
            "lat": np.array([4.0]),
            "n_obs": np.array([1000], dtype=np.int32),
            "R": np.array([1.0], dtype=np.float32),
            "p_R": np.array([0.0], dtype=np.float32),
            "rho": np.array([1.0], dtype=np.float32),
            "p_rho": np.array([0.0], dtype=np.float32),
            "RMSD": np.array([0.0], dtype=np.float32),
            "BIAS": np.array([0.0], dtype=np.float32),
            "tau": np.array([np.nan], dtype=np.float32),
            "p_tau": np.array([np.nan], dtype=np.float32),
        },
        (("DS2", "y"), ("DS3", "y")): {
            "gpi": np.array([4], dtype=np.int32),
            "lon": np.array([4.0]),
            "lat": np.array([4.0]),
            "n_obs": np.array([1000], dtype=np.int32),
            "R": np.array([1.0], dtype=np.float32),
            "p_R": np.array([0.0], dtype=np.float32),
            "rho": np.array([1.0], dtype=np.float32),
            "p_rho": np.array([0.0], dtype=np.float32),
            "RMSD": np.array([0.0], dtype=np.float32),
            "BIAS": np.array([0.0], dtype=np.float32),
            "tau": np.array([np.nan], dtype=np.float32),
            "p_tau": np.array([np.nan], dtype=np.float32),
        },
    }

    datasets = setup_TestDatasets()
    dm = DataManager(
        datasets,
        "DS1",
        read_ts_names={d: "read" for d in ["DS1", "DS2", "DS3"]},
    )

    process = Validation(
        dm,
        "DS1",
        temporal_matcher=temporal_matchers.BasicTemporalMatching(
            window=1 / 24.0).combinatory_matcher,
        scaling="cdf_match",
        metrics_calculators={
            (3, 2):
                metrics_calculators.BasicMetrics(other_name="k1").calc_metrics
        },
    )

    jobs = process.get_processing_jobs()
    for job in jobs:
        results = process.calc(*job)
        assert sorted(list(results)) == sorted(list(tst_results))


def test_validation_n3_k2_temporal_matching_no_matches2():

    empty_result = {
        'gpi': np.array([4], dtype=np.int32),
        'lon': np.array([4.]),
        'lat': np.array([4.]),
        'status': np.array([eh.NO_TEMP_MATCHED_DATA], dtype=np.int32),
        'n_obs': np.array([0], dtype=np.int32),
        'R': np.array([np.nan], dtype=np.float32),
        'p_R': np.array([np.nan], dtype=np.float32),
        'rho': np.array([np.nan], dtype=np.float32),
        'p_rho': np.array([np.nan], dtype=np.float32),
        'RMSD': np.array([np.nan], dtype=np.float32),
        'BIAS': np.array([np.nan], dtype=np.float32),
        'tau': np.array([np.nan], dtype=np.float32),
        'p_tau': np.array([np.nan], dtype=np.float32)
    }

    # only DS1 and DS3 overlap, so we only get two valid results
    tst_results = {
        (("DS1", "x"), ("DS2", "y")): empty_result,
        (("DS1", "x"), ("DS3", "x")): {
            "n_obs": np.array([1000], dtype=np.int32),
            "tau": np.array([np.nan], dtype=np.float32),
            "gpi": np.array([4], dtype=np.int32),
            "RMSD": np.array([0.0], dtype=np.float32),
            "lon": np.array([4.0]),
            "p_tau": np.array([np.nan], dtype=np.float32),
            "BIAS": np.array([0.0], dtype=np.float32),
            "p_rho": np.array([0.0], dtype=np.float32),
            "rho": np.array([1.0], dtype=np.float32),
            "lat": np.array([4.0]),
            "R": np.array([1.0], dtype=np.float32),
            "p_R": np.array([0.0], dtype=np.float32),
        },
        (("DS1", "x"), ("DS3", "y")): {
            "n_obs": np.array([1000], dtype=np.int32),
            "tau": np.array([np.nan], dtype=np.float32),
            "gpi": np.array([4], dtype=np.int32),
            "RMSD": np.array([0.0], dtype=np.float32),
            "lon": np.array([4.0]),
            "p_tau": np.array([np.nan], dtype=np.float32),
            "BIAS": np.array([0.0], dtype=np.float32),
            "p_rho": np.array([0.0], dtype=np.float32),
            "rho": np.array([1.0], dtype=np.float32),
            "lat": np.array([4.0]),
            "R": np.array([1.0], dtype=np.float32),
            "p_R": np.array([0.0], dtype=np.float32),
        },
        (("DS2", "y"), ("DS3", "x")): empty_result,
        (("DS2", "y"), ("DS3", "y")): empty_result,
    }

    datasets = setup_three_with_two_overlapping()
    dm = DataManager(
        datasets,
        "DS1",
        read_ts_names={d: "read" for d in ["DS1", "DS2", "DS3"]},
    )

    process = Validation(
        dm,
        "DS1",
        temporal_matcher=temporal_matchers.BasicTemporalMatching(
            window=1 / 24.0).combinatory_matcher,
        scaling="cdf_match",
        metrics_calculators={
            (3, 2):
                metrics_calculators.BasicMetrics(other_name="k1").calc_metrics
        },
    )

    jobs = process.get_processing_jobs()
    for job in jobs:
        results = process.calc(*job, handle_errors="ignore")
        assert sorted(list(results)) == sorted(list(tst_results))

    jobs = process.get_processing_jobs()
    for job in jobs:
        with pytest.raises(eh.NoTempMatchedDataError):
            results = process.calc(*job, handle_errors="raise")


def test_validation_n3_k2_masking_no_data_remains():
    datasets = setup_TestDatasets()

    # setup masking datasets

    grid = grids.CellGrid(
        np.array([1, 2, 3, 4]),
        np.array([1, 2, 3, 4]),
        np.array([4, 4, 2, 1]),
        gpis=np.array([1, 2, 3, 4]),
    )

    mds1 = GriddedTsBase("", grid, MaskingTestDataset)
    mds2 = GriddedTsBase("", grid, MaskingTestDataset)

    mds = {
        "masking1": {
            "class": mds1,
            "columns": ["x"],
            "args": [],
            "kwargs": {
                "limit": 500
            },
            "use_lut": False,
            "grids_compatible": True,
        },
        "masking2": {
            "class": mds2,
            "columns": ["x"],
            "args": [],
            "kwargs": {
                "limit": 1000
            },
            "use_lut": False,
            "grids_compatible": True,
        },
    }

    process = Validation(
        datasets,
        "DS1",
        temporal_matcher=temporal_matchers.BasicTemporalMatching(
            window=1 / 24.0).combinatory_matcher,
        scaling="cdf_match",
        metrics_calculators={
            (3, 2):
                metrics_calculators.BasicMetrics(other_name="k1").calc_metrics
        },
        masking_datasets=mds,
    )

    gpi_info = (1, 1, 1)
    ref_df = datasets["DS1"]["class"].read(1)
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=DeprecationWarning)
        new_ref_df = process.mask_dataset(ref_df, gpi_info)
    assert len(new_ref_df) == 0
    nptest.assert_allclose(new_ref_df.x.values, np.arange(1000, 1000))
    jobs = process.get_processing_jobs()
    for job in jobs:
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=DeprecationWarning)
            results = process.calc(*job)
        tst = []
        assert sorted(list(results)) == sorted(list(tst))
        for key, tst_key in zip(sorted(results), sorted(tst)):
            nptest.assert_almost_equal(results[key]["n_obs"],
                                       tst[tst_key]["n_obs"])


def test_validation_n3_k2_masking():
    # test result for one gpi in a cell
    tst_results_one = {
        (("DS1", "x"), ("DS3", "y")): {
            "n_obs": np.array([250], dtype=np.int32)
        },
        (("DS1", "x"), ("DS2", "y")): {
            "n_obs": np.array([250], dtype=np.int32)
        },
        (("DS1", "x"), ("DS3", "x")): {
            "n_obs": np.array([250], dtype=np.int32)
        },
        (("DS2", "y"), ("DS3", "x")): {
            "n_obs": np.array([250], dtype=np.int32)
        },
        (("DS2", "y"), ("DS3", "y")): {
            "n_obs": np.array([250], dtype=np.int32)
        },
    }

    # test result for two gpis in a cell
    tst_results_two = {
        (("DS1", "x"), ("DS3", "y")): {
            "n_obs": np.array([250, 250], dtype=np.int32)
        },
        (("DS1", "x"), ("DS2", "y")): {
            "n_obs": np.array([250, 250], dtype=np.int32)
        },
        (("DS1", "x"), ("DS3", "x")): {
            "n_obs": np.array([250, 250], dtype=np.int32)
        },
        (("DS2", "y"), ("DS3", "x")): {
            "n_obs": np.array([250, 250], dtype=np.int32)
        },
        (("DS2", "y"), ("DS3", "y")): {
            "n_obs": np.array([250, 250], dtype=np.int32)
        },
    }

    # cell 4 in this example has two gpis so it returns different results.
    tst_results = {1: tst_results_one, 1: tst_results_one, 2: tst_results_two}

    datasets = setup_TestDatasets()

    # setup masking datasets

    grid = grids.CellGrid(
        np.array([1, 2, 3, 4]),
        np.array([1, 2, 3, 4]),
        np.array([4, 4, 2, 1]),
        gpis=np.array([1, 2, 3, 4]),
    )

    mds1 = GriddedTsBase("", grid, MaskingTestDataset)
    mds2 = GriddedTsBase("", grid, MaskingTestDataset)

    mds = {
        "masking1": {
            "class": mds1,
            "columns": ["x"],
            "args": [],
            "kwargs": {
                "limit": 500
            },
            "use_lut": False,
            "grids_compatible": True,
        },
        "masking2": {
            "class": mds2,
            "columns": ["x"],
            "args": [],
            "kwargs": {
                "limit": 750
            },
            "use_lut": False,
            "grids_compatible": True,
        },
    }

    process = Validation(
        datasets,
        "DS1",
        temporal_matcher=temporal_matchers.BasicTemporalMatching(
            window=1 / 24.0).combinatory_matcher,
        scaling="cdf_match",
        metrics_calculators={
            (3, 2):
                metrics_calculators.BasicMetrics(other_name="k1").calc_metrics
        },
        masking_datasets=mds,
    )

    gpi_info = (1, 1, 1)
    ref_df = datasets["DS1"]["class"].read(1)
    with warnings.catch_warnings():
        warnings.simplefilter(
            "ignore", category=DeprecationWarning
        )  # read_ts is hard coded when using mask_data
        new_ref_df = process.mask_dataset(ref_df, gpi_info)
    assert len(new_ref_df) == 250
    nptest.assert_allclose(new_ref_df.x.values, np.arange(750, 1000))
    jobs = process.get_processing_jobs()
    for job in jobs:

        with warnings.catch_warnings():
            # most warnings here are caused by the read_ts function that cannot
            # be changed when using a masking data set
            warnings.simplefilter("ignore", category=DeprecationWarning)
            results = process.calc(*job)

        tst = tst_results[len(job[0])]
        assert sorted(list(results)) == sorted(list(tst))
        for key, tst_key in zip(sorted(results), sorted(tst)):
            nptest.assert_almost_equal(results[key]["n_obs"],
                                       tst[tst_key]["n_obs"])


@pytest.mark.slow
@pytest.mark.full_framework
def test_ascat_ismn_validation_metadata_rolling(ascat_reader, ismn_reader):
    """
    Test processing framework with some ISMN and ASCAT sample data
    """
    jobs = []

    ids = ismn_reader.get_dataset_ids(
        variable="soil_moisture", min_depth=0, max_depth=0.1)

    metadata_dict_template = {
        "network": np.array(["None"], dtype="U256"),
        "station": np.array(["None"], dtype="U256"),
        "landcover": np.float32([np.nan]),
        "climate": np.array(["None"], dtype="U4"),
    }

    for idx in ids:
        metadata = ismn_reader.read_metadata(idx, format='obj')
        metadata_dict = [{
            "network": metadata["network"][1],
            "station": metadata["station"][1],
            "landcover": metadata["lc_2010"][1],
            "climate": metadata["climate_KG"][1],
        }]
        jobs.append((idx, metadata["longitude"][1], metadata["latitude"][1],
                     metadata_dict))

    save_path = tempfile.mkdtemp()

    # Create the validation object.

    datasets = {
        "ISMN": {
            "class": ismn_reader,
            "columns": ["soil_moisture"]
        },
        "ASCAT": {
            "class": ascat_reader,
            "columns": ["sm"],
            "kwargs": {
                "mask_frozen_prob": 80,
                "mask_snow_prob": 80,
                "mask_ssf": True,
            },
        },
    }

    read_ts_names = {"ASCAT": "read", "ISMN": "read"}
    period = [datetime(2007, 1, 1), datetime(2014, 12, 31)]

    datasets = DataManager(
        datasets, "ISMN", period, read_ts_names=read_ts_names)

    process = Validation(
        datasets,
        "ISMN",
        temporal_ref="ASCAT",
        scaling="cdf_match",
        scaling_ref="ASCAT",
        metrics_calculators={
            (2, 2):
                metrics_calculators.RollingMetrics(
                    other_name="k1",
                    metadata_template=metadata_dict_template).calc_metrics
        },
        period=period,
    )

    for job in jobs:
        results = process.calc(*job)
        netcdf_results_manager(
            results, save_path, ts_vars=["R", "p_R", "RMSD"])

    results_fname = os.path.join(save_path,
                                 "ASCAT.sm_with_ISMN.soil_moisture.nc")

    target_vars = {
        "network":
            np.array(
                [
                    "MAQU",
                    "MAQU",
                    "SCAN",
                    "SCAN",
                    "SCAN",
                    "SOILSCAPE",
                    "SOILSCAPE",
                    "SOILSCAPE",
                ],
                dtype="U256",
            )
    }
    vars_should = [
        u"gpi", u"RMSD", u"lon", u"lat", u"R", u"p_R", u"time", u"idx",
        u"_row_size", u"status"
    ]
    for key, value in metadata_dict_template.items():
        vars_should.append(key)

    check_results(
        filename=results_fname, target_vars=target_vars, variables=vars_should)

    reader = PointDataResults(results_fname, read_only=True)
    df = reader.read_loc(None)
    assert np.all(df.gpi.values == np.arange(8))
    assert reader.read_ts(0).index.size == 357
    assert np.all(
        reader.read_ts(1).columns.values == np.array(["R", "p_R", "RMSD"]))

    ascat_reader.close()


def test_args_to_iterable_non_iterables():
    gpis = 1
    lons = 1
    lats = 1
    arg1 = 1
    arg2 = 2
    arg3 = 3
    gpis_, lons_, lats_, args = args_to_iterable(
        gpis, lons, lats, arg1, arg2, arg3, n=3)

    assert gpis_ == [gpis]
    assert lons_ == [lons]
    assert lats_ == [lats]
    assert args == ([arg1], [arg2], [arg3])


def test_args_to_iterable_n3():
    gpis = [1, 2, 3]
    lons = [2, 3, 4]
    lats = [3, 4, 5]
    arg1 = [1, 1, 1]
    arg2 = [1, 1, 1]
    gpis_, lons_, lats_, args = args_to_iterable(
        gpis, lons, lats, arg1, arg2, n=3)

    assert gpis_ == gpis
    assert lons_ == lons
    assert lats_ == lats
    assert args == (arg1, arg2)

    zipped_should = [(1, 2, 3, 1, 1), (2, 3, 4, 1, 1), (3, 4, 5, 1, 1)]

    for i, t in enumerate(zip(gpis_, lons_, lats_, *args)):
        assert zipped_should[i] == t


def test_args_to_iterable_mixed():
    gpis = [1, 2, 3]
    lons = [2, 3, 4]
    lats = 1
    arg1 = 1
    gpis_, lons_, lats_, args = args_to_iterable(gpis, lons, lats, arg1)

    assert gpis_ == gpis
    assert lons_ == lons
    assert lats_ == [lats]
    assert args == [arg1]


def test_args_to_iterable_mixed_strings():
    gpis = [1, 2, 3]
    lons = [2, 3, 4]
    lats = 1
    arg1 = "test"
    gpis_, lons_, lats_, args = args_to_iterable(gpis, lons, lats, arg1)

    assert gpis_ == gpis
    assert lons_ == lons
    assert lats_ == [lats]
    assert args == [arg1]


#######################################################################
# Tests for new temporal matcher & metric calculators
#######################################################################


def test_combined_matching_scaling():
    n_datasets = 5
    npoints = 5
    nsamples = 100

    datasets = create_datasets(n_datasets, npoints, nsamples)

    metric_calculator = PairwiseIntercomparisonMetrics()

    val = Validation(
        datasets,
        spatial_ref="0-ERA5",
        metrics_calculators={(n_datasets, 2): metric_calculator.calc_metrics},
        temporal_matcher=make_combined_temporal_matcher(pd.Timedelta(12, "H")),
        scaling="mean_std",
        scaling_ref="0-ERA5",
    )
    gpis = list(range(npoints))
    val.calc(gpis, gpis, gpis, rename_cols=False, only_with_reference=True)


if __name__ == "__main__":
    # for profiling with cProfile, on the command line run
    # python -m cProfile -o ascat_ismn_validation.profile test_validation.py
    # output can be investigated with snakeviz
    test_ascat_ismn_validation()
    # profiling result: most of the time is spend loading the data (110s), next
    # bigger chunk of computation time is temporal matching (6.7s, old
    # implementation, 0.3s new implementation)



# -*- coding: utf-8 -*-

from setuptools import setup
from setuptools.command.build_ext import build_ext as _build_ext
from setuptools.command.sdist import sdist as _sdist
from setuptools.extension import Extension
import numpy


# list of C/Cython extensions
def get_ext_modules(ext):
    return [
        Extension(
            "pytesmo.time_series.filters",
            ["src/pytesmo/time_series/filters" + ext],
            include_dirs=[numpy.get_include()],
            define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
        ),
        Extension(
            "pytesmo.metrics._fast",
            ["src/pytesmo/metrics/_fast" + ext],
            include_dirs=[numpy.get_include()],
            define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
        ),
        Extension(
            "pytesmo.metrics._fast_pairwise",
            ["src/pytesmo/metrics/_fast_pairwise" + ext],
            include_dirs=[numpy.get_include()],
            define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
        ),
    ]


# defining a custom cythonize function that sets all the options we want and
# that can be reused for the different options
def cythonize_extensions():
    from Cython.Build import cythonize

    cythonize(
        get_ext_modules(".pyx"),
        compiler_directives={
            "embedsignature": True,
            "language_level": 3,
            # "warn.undeclared": True,
            # "warn.maybe_unitialized": True,
            "warn.unused": True,
            # "warn.unused_arg": True,
            # "warn.unused_result": True,
        },
        # include_path=[numpy.get_include()],
    )


class CythonizeMixin(object):

    user_options = [
        ("cythonize", None, "recreate the c extionsions with cython")
    ]

    def initialize_options(self):
        super().initialize_options()
        self.cythonize = False

    def run(self):
        if self.cythonize:
            cythonize_extensions()
        super().run()


class sdist(CythonizeMixin, _sdist):
    user_options = getattr(_sdist, 'user_options', [])\
               + CythonizeMixin.user_options


class build_ext(CythonizeMixin, _build_ext):
    user_options = getattr(_build_ext, 'user_options', [])\
               + CythonizeMixin.user_options


if __name__ == "__main__":
    cmdclass = {}
    cmdclass["sdist"] = sdist
    cmdclass["build_ext"] = build_ext
    setup(
        cmdclass=cmdclass,
        # at this point the C modules have already been generated if necessary
        ext_modules=get_ext_modules(".c"),
        use_scm_version={"version_scheme": "no-guess-dev"}
    )



try:
    from itertools import izip as zip
except ImportError:
    # python 3
    pass

import numpy as np
import pandas as pd
from pygeogrids.grids import CellGrid
import warnings
import logging
from typing import Mapping, Tuple, List

from pytesmo.validation_framework.data_manager import DataManager
from pytesmo.validation_framework.data_manager import get_result_names
from pytesmo.validation_framework.data_manager import get_result_combinations
from pytesmo.validation_framework.data_scalers import DefaultScaler
import pytesmo.validation_framework.temporal_matchers as temporal_matchers
from pytesmo.utils import ensure_iterable
from distutils.version import LooseVersion

import pytesmo.validation_framework.error_handling as eh


class Validation(object):

    """
    Class for the validation process.

    Parameters
    ----------
    datasets : dict of dicts or DataManager
        :Keys: string, datasets names
        :Values: dict, containing the following fields
        :py:class:`pytesmo.validation_framework.data_manager.DataManager`

            'class': object
                Class containing the method read_ts for reading the data.
            'columns': list
                List of columns which will be used in the validation process.
            'args': list, optional
                Args for reading the data.
            'kwargs': dict, optional
                Kwargs for reading the data
            'grids_compatible': boolean, optional
                If set to True the grid point index is used directly when
                reading other, if False then lon, lat is used and a nearest
                neighbour search is necessary.
            'use_lut': boolean, optional
                If set to True the grid point index (obtained from a
                calculated lut between reference and other) is used when
                reading other, if False then lon, lat is used and a
                nearest neighbour search is necessary.
            'lut_max_dist': float, optional
                Maximum allowed distance in meters for the lut calculation.
    spatial_ref: string
        Name of the dataset used as a spatial, temporal and scaling reference.
        temporal and scaling references can be changed if needed. See the
        optional parameters ``temporal_ref`` and ``scaling_ref``.
    metrics_calculators : dict of functions
        The keys of the dict are tuples with the following structure: (n, k)
        with n >= 2 and n>=k. n must be equal to the number of datasets now.
        n is the number of datasets that should be temporally matched to the
        reference dataset and k is how many columns the metric calculator
        will get at once.
        What this means is that it is e.g. possible to temporally match 3
        datasets with 3 columns in total and then give the combinations of
        these columns to the metric calculator in sets of 2 by specifying the
        dictionary like:

        .. code::

            { (3, 2): metric_calculator}

        The values are functions that take an input DataFrame with the columns
        'ref' for the reference and 'n1', 'n2' and
        so on for other datasets as well as a dictionary mapping the column
        names to the names of the original datasets. In this way multiple
        metric calculators can be applied to different combinations of n
        input datasets.
    temporal_matcher: function, optional
        function that takes a dict of dataframes and a reference_key.
        It performs the temporal matching on the data and returns a dictionary
        of matched DataFrames that should be evaluated together by the metric
        calculator.
    temporal_window: float, optional
        Window to allow in temporal matching in days. The window is allowed
        on both sides of the timestamp of the temporal reference data.
        Only used with the standard temporal matcher.
    temporal_ref: string, optional
        If the temporal matching should use another dataset than the spatial
        reference as a reference dataset then give the dataset name here.
    period : list, optional
        Of type [datetime start, datetime end]. If given then the two input
        datasets will be truncated to start <= dates <= end.
    masking_datasets : dict of dictionaries
        Same format as the datasets with the difference that the `read`
        method of these datasets has to return pandas.DataFrames with only
        boolean columns. True means that the observations at this timestamp
        should be masked and False means that it should be kept.
    scaling : string, None or class instance
        - If set then the data will be scaled into the reference space using
          the method specified by the string using the
          :py:class:`pytesmo.validation_framework.data_scalers.DefaultScaler`
          class.
        - If set to None then no scaling will be performed.
        - It can also be set to a class instance that implements a
          ``scale(self, data, reference_index, gpi_info)`` method. See
          :py:class:`pytesmo.validation_framework.data_scalers.DefaultScaler`
          for an example.
    scaling_ref : string, optional
        If the scaling should be done to another dataset than the spatial
        reference then give the dataset name here.

    Methods
    -------
    calc(job)
        Takes either a cell or a gpi_info tuple and performs the validation.
    get_processing_jobs()
        Returns processing jobs that this process can understand.
    """

    def __init__(
        self,
        datasets,
        spatial_ref,
        metrics_calculators,
        temporal_matcher=None,
        temporal_window=1 / 24.0,
        temporal_ref=None,
        masking_datasets=None,
        period=None,
        scaling="cdf_match",
        scaling_ref=None,
    ):

        if isinstance(datasets, DataManager):
            self.data_manager = datasets
        else:
            self.data_manager = DataManager(datasets, spatial_ref, period)

        self.temp_matching = temporal_matcher
        if self.temp_matching is None:
            warnings.warn(
                "You are using the default temporal matcher. If you are using "
                "one of the newer metric calculators "
                "(PairwiseIntercomparisonMetrics, TripleCollocationMetrics) "
                "you should probably use `make_combined_temporal_matcher`"
                " instead. Have a look at the documentation of the metric "
                "calculators for more info."
            )
            self.temp_matching = temporal_matchers.BasicTemporalMatching(
                window=temporal_window
            ).combinatory_matcher

        self.temporal_ref = temporal_ref
        if self.temporal_ref is None:
            self.temporal_ref = self.data_manager.reference_name

        self.metrics_c = metrics_calculators
        for n, k in self.metrics_c:
            if n < len(self.data_manager.datasets.keys()):
                raise ValueError("n must be equal to the number of datasets")

        self.masking_dm = None
        if masking_datasets is not None:
            # add temporal reference dataset to the masking datasets since it
            # is necessary for temporally matching the masking datasets to the
            # common time stamps. Use _reference here to make a clash with the
            # names of the masking datasets unlikely
            masking_datasets.update(
                {"_reference": datasets[self.temporal_ref]}
            )
            self.masking_dm = DataManager(
                masking_datasets, "_reference", period=period
            )

        if type(scaling) == str:
            self.scaling = DefaultScaler(scaling)
        else:
            self.scaling = scaling
        self.scaling_ref = scaling_ref
        if self.scaling_ref is None:
            self.scaling_ref = self.data_manager.reference_name

        self.luts = self.data_manager.get_luts()

    def calc(
        self,
        gpis,
        lons,
        lats,
        *args,
        rename_cols=True,
        only_with_reference=False,
        handle_errors='raise',
    ) -> Mapping[Tuple[str], Mapping[str, np.ndarray]]:
        """
        The argument iterables (lists or numpy.ndarrays) are processed one
        after the other in tuples of the form (gpis[n], lons[n], lats[n],
        arg1[n], ..).

        Parameters
        ----------
        gpis: iterable
            The grid point indices is an identificator by which the
            spatial reference dataset can be read. This is either a list
            or a numpy.ndarray or any other iterable containing this indicator.
        lons: iterable
            Longitudes of the points identified by the gpis. Has to be the same
            size as gpis.
        lats: iterable
            latitudes of the points identified by the gpis. Has to be the same
            size as gpis.
        args: iterables
            any addiational arguments have to have the same size as the gpis
            iterable. They are given to the metrics calculators as
            metadata. Common usage is e.g. the long name or network name of an
            in situ station.
        rename_cols : bool, optional
            Whether to rename the columns to "ref", "k1", ... before passing
            the dataframe to the metrics calculators. Default is True.
        only_with_reference : bool, optional
            If this is enabled, only combinations that include the reference
            dataset (from the data manager) are calculated.
        handle_errors: str, optional (default: 'raise')
            Governs how to handle errors::

            * `raise`: If an error occurs during validation, raise exception.
            * `ignore`: If an error occurs, assign the correct return code
              to the result template and continue with the next GPI.

        Returns
        -------
        compact_results : dict of dicts
            :Keys: result names, combinations of
                  (referenceDataset.column, otherDataset.column)
            :Values: dict containing the elements returned
                  by metrics_calculator

        """
        handle_errors = handle_errors.lower()
        error_handling_options = ["raise", "ignore"]
        assert handle_errors in error_handling_options, (
            f"'handle_errors' must be one of {error_handling_options}"
        )

        if len(args) > 0:
            gpis, lons, lats, args = args_to_iterable(
                gpis, lons, lats, *args, n=3
            )
        else:
            gpis, lons, lats = args_to_iterable(gpis, lons, lats)

        results = {}
        for gpi_info in zip(gpis, lons, lats, *args):

            try:
                try:
                    df_dict = self.data_manager.get_data(
                        gpi_info[0], gpi_info[1], gpi_info[2]
                    )
                except Exception as e:
                    raise eh.DataManagerError(
                        f"Getting the data for gpi {gpi_info} failed!")

                # if no data is available continue with the next gpi
                if len(df_dict) == 0:
                    raise eh.NoGpiDataError(f"No data for gpi {gpi_info}")
                matched_data, result, used_data = self.perform_validation(
                    df_dict,
                    gpi_info,
                    rename_cols=rename_cols,
                    only_with_reference=only_with_reference,
                    handle_errors=handle_errors,
                )
            except Exception as e:
                if handle_errors == 'raise':
                    raise e
                elif handle_errors == "ignore":
                    logging.error(f"{gpi_info}: {e}")
                    result = self.dummy_validation_result(
                        gpi_info, rename_cols=rename_cols,
                        only_with_reference=only_with_reference)
                    if isinstance(e, eh.ValidationError):
                        retcode = e.return_code
                    else:
                        retcode = eh.VALIDATION_FAILED
                    for key in result:
                        result[key][0]["status"][0] = retcode

            # add result of one gpi to global results dictionary
            for r in result:
                if r not in results:
                    results[r] = []
                results[r] = results[r] + result[r]

        # So far, results is a dictionary mapping from a validation name/key
        # (the involved datasets) to lists of individual result dictionaries
        # for each gpi.
        # Here, these lists are summarized to have a single dictionary of numpy
        # arrays for each validation key
        compact_results = {}
        for key in results.keys():
            compact_results[key] = {}
            for field_name in results[key][0].keys():
                entries = []
                for result in results[key]:
                    entries.append(result[field_name][0])
                compact_results[key][field_name] = np.array(
                    entries, dtype=results[key][0][field_name].dtype
                )

        return compact_results

    def perform_validation(
        self,
        df_dict,
        gpi_info,
        rename_cols=True,
        only_with_reference=False,
        handle_errors="raise",
    ) -> Mapping[Tuple[str], List[Mapping[str, np.ndarray]]]:
        """
        Perform the validation for one grid point index and return the
        matched datasets as well as the calculated metrics.

        Parameters
        ----------
        df_dict: dict of pandas.DataFrames
            DataFrames read by the data readers for each dataset
        gpi_info: tuple
            tuple of at least, (gpi, lon, lat)
        rename_cols : bool, optional
            Whether to rename the columns to "ref", "k1", ... before passing
            the dataframe to the metrics calculators. Default is True.
        only_with_reference: bool, optional (default: False)
            Only compute metrics for dataset combinations where the reference
            is included.

        Returns
        -------
        matched_n: dict of pandas.DataFrames
            temporally matched data stored by (n, k) tuples
        results: dict
            Dictonary of calculated metrics stored by dataset combinations
            tuples.
        used_data: dict
            The DataFrame used for calculation of each set of metrics.

        Raises
        ------
        eh.TemporalMatchingError :
            If temporal matching failed
        eh.NoTempMatchedDataError :
            If there is insufficient data or the temporal matching did not
            return data.
        eh.ScalingError :
            If scaling failed
        """
        results = {}
        used_data = {}
        matched_n = {}

        if self.masking_dm is not None:
            ref_df = df_dict[self.temporal_ref]
            masked_ref_df = self.mask_dataset(ref_df, gpi_info)
            if len(masked_ref_df) == 0:
                return matched_n, results, used_data

            df_dict[self.temporal_ref] = masked_ref_df

        # we only need the data columns
        data_df_dict = {}
        for ds in df_dict:
            columns = self.data_manager.datasets[ds]["columns"]
            data_df_dict[ds] = df_dict[ds][columns]

        # matched_n is a dictionary mapping from dataset combinations as keys
        # to pandas dataframes
        try:
            matched_n = self.temporal_match_datasets(data_df_dict)
        except Exception:
            raise eh.TemporalMatchingError(
                f"Temporal matching failed for gpi {gpi_info}!"
            )

        for n, k in self.metrics_c:
            metrics_calculator = self.metrics_c[(n, k)]

            def dummy_result():
                # to get only an empty template returned, we pass an
                # empty dataframe here, with the correct column names
                dummy_df = pd.DataFrame([], columns=result_ds_names)
                metrics = metrics_calculator(dummy_df, gpi_info)
                return metrics

            n_matched_data = matched_n[(n, k)]
            if len(n_matched_data) == 0:
                # this would happen if self.temp_matching returns an empty list
                # or dictionary for n=n, k=k
                raise eh.NoTempMatchedDataError(
                    f"No temporally matched data for ({n}, {k})"
                    f" and metric calculator {self.metrics_c[(n,k)]}"
                    f" for gpi {gpi_info}!"
                )
            result_names = get_result_combinations(
                self.data_manager.ds_dict, n=k
            )
            for data, result_key in self.k_datasets_from(
                n_matched_data, result_names
            ):

                # it might also be a good idea to move this to
                # `get_result_combinations`
                result_ds_names = [key[0] for key in result_key]
                if only_with_reference:
                    if self.data_manager.reference_name not in result_ds_names:
                        continue

                if result_key not in results.keys():
                    results[result_key] = []

                if len(data) == 0:
                    if handle_errors == "raise":
                        raise eh.NoTempMatchedDataError(
                            f"Temporal matching resulted in empty dataset for"
                            f" {result_key} for gpi {gpi_info}"
                        )
                    else:
                        metrics = dummy_result()
                        metrics["status"][0] = eh.NO_TEMP_MATCHED_DATA
                        results[result_key].append(metrics)
                        continue

                # at this stage we can drop the column multiindex and just use
                # the dataset name
                if LooseVersion(pd.__version__) < LooseVersion("0.23"):
                    data.columns = data.columns.droplevel(level=1)
                else:
                    data = data.rename(columns=lambda x: x[0])

                if self.scaling is not None:
                    # get scaling index by finding the column in the
                    # DataFrame that belongs to the scaling reference
                    scaling_index = data.columns.tolist().index(
                        self.scaling_ref
                    )
                    try:
                        data = self.scaling.scale(
                            data, scaling_index, gpi_info
                        )
                    except Exception as e:
                        raise eh.ScalingError(
                            f"Scaling failed for {result_key} for gpi"
                            f" {gpi_info}!"
                        )

                    # Drop the scaling reference if it was not in the intended
                    # results
                    if self.scaling_ref not in [key[0] for key in result_key]:
                        data = data.drop(columns=[self.scaling_ref])

                # Rename the columns to 'ref', 'k1', 'k2', ...
                if rename_cols:
                    rename_dict = {}
                    for i, r in enumerate(result_key):
                        rename_dict[r[0]] = f"k{i}" if i > 0 else "ref"
                    data.rename(columns=rename_dict, inplace=True)

                used_data[result_key] = data
                try:
                    metrics = metrics_calculator(data, gpi_info)
                except Exception:
                    if handle_errors == "raise":
                        raise eh.MetricsCalculationError(
                            f"Metrics calculation failed for {result_key}"
                            f" for gpi {gpi_info}!"
                        )
                    else:
                        metrics = dummy_result()
                        metrics["status"][0] = eh.METRICS_CALCULATION_FAILED
                results[result_key].append(metrics)

        return matched_n, results, used_data

    def dummy_validation_result(
        self,
        gpi_info,
        rename_cols=True,
        only_with_reference=False,
    ) -> Mapping[Tuple[str], List[Mapping[str, np.ndarray]]]:
        """
        Creates an empty result dictionary to be used if perform_validation
        fails
        """
        results = {}
        for n, k in self.metrics_c:
            result_names = get_result_combinations(
                self.data_manager.ds_dict, n=k
            )
            for result_key in result_names:
                # it might also be a good idea to move this to
                # `get_result_combinations`
                result_ds_names = [key[0] for key in result_key]
                if only_with_reference:
                    if self.data_manager.reference_name not in result_ds_names:
                        continue
                metrics_calculator = self.metrics_c[(n, k)]
                # to get only an empty template returned, we pass an empty
                # dataframe here, with the correct column names
                dummy_df = pd.DataFrame([], columns=result_ds_names)
                metrics = metrics_calculator(dummy_df, gpi_info)
                if result_key not in results.keys():
                    results[result_key] = []
                results[result_key].append(metrics)
        return results

    def mask_dataset(self, ref_df, gpi_info):
        """
        Mask the temporal reference dataset with the data read
        through the masking datasets.

        Parameters
        ----------
        gpi_info: tuple
            tuple of at least, (gpi, lon, lat)

        Returns
        -------
        mask: numpy.ndarray
            boolean array of the size of the temporal reference read
        """

        matched_masking = self.temporal_match_masking_data(ref_df, gpi_info)
        # this will only be one element since n is the same as the
        # number of masking datasets
        result_names = get_result_names(
            self.masking_dm.ds_dict, "_reference", n=2
        )
        choose_all = pd.DataFrame(index=ref_df.index)
        for data, result in self.k_datasets_from(
            matched_masking, result_names, include_scaling_ref=False
        ):
            if len(data) == 0:
                continue

            for key in result:
                if key[0] != "_reference":
                    # this is necessary since the boolean datatype might have
                    # been changed to float 1.0 and 0.0 issue with temporal
                    # resampling that is not easily resolved since most
                    # datatypes have no nan representation.
                    choose = pd.Series((data[key] is False), index=data.index)
                    choose = choose.reindex(
                        index=choose_all.index, fill_value=True
                    )
                    choose_all[key] = choose.copy()
        choosing = choose_all.apply(np.all, axis=1)

        return ref_df[choosing]

    def temporal_match_masking_data(self, ref_df, gpi_info):
        """
        Temporal match the masking data to the reference DataFrame

        Parameters
        ----------
        ref_df: pandas.DataFrame
            Reference data
        gpi_info: tuple or list
            contains, (gpi, lon, lat)

        Returns
        -------
        matched_masking: dict of pandas.DataFrames
            Contains temporally matched masking data. This dict has only one
            key being a tuple that contains the matched datasets.
        """

        # read only masking datasets and use the already read reference
        masking_df_dict = self.masking_dm.get_other_data(
            gpi_info[0], gpi_info[1], gpi_info[2]
        )
        masking_df_dict.update({"_reference": ref_df})
        matched_masking = self.temp_matching(
            masking_df_dict, "_reference", n=2
        )
        return matched_masking

    def temporal_match_datasets(self, df_dict):
        """
        Temporally match all the requested combinations of datasets.

        Parameters
        ----------
        df_dict: dict of pandas.DataFrames
            DataFrames read by the data readers for each dataset

        Returns
        -------
        matched_n: dict of pandas.DataFrames
            for each (n, k) in the metrics calculators the n temporally
            matched dataframes
        """

        matched_n = {}
        for n, k in self.metrics_c:
            matched_data = self.temp_matching(
                df_dict, self.temporal_ref, n=n, k=k
            )

            matched_n[(n, k)] = matched_data

        return matched_n

    def k_datasets_from(
        self, n_matched_data, result_names, include_scaling_ref=True
    ):
        """
        Extract k datasets from n temporally matched ones.

        This is used to send combinations of k datasets to
        metrics calculators expecting only k datasets.

        Parameters
        ----------
        n_matched_data: dict of pandas.DataFrames
            DataFrames in which n datasets were temporally matched.
            The key is a tuple of the dataset names.
        result_names: list
            result names to extract
        include_scaling_ref: boolean, optional
            if set the scaling reference will always be included.
            Should only be disabled for getting the masking datasets

        Yields
        ------
        data: pd.DataFrame
            pandas DataFrame with k columns extracted from the
            temporally matched datasets
        result: tuple
            Tuple describing which datasets and columns are in
            the returned data. ((dataset_name, column_name),
            (dataset_name2, column_name2))
        """

        for result in result_names:
            result_extract = result
            if self.scaling is not None and include_scaling_ref:
                # always make sure the scaling reference is included in the
                # results otherwise the scaling will fail
                scaling_ref_column = self.data_manager.datasets[
                    self.scaling_ref
                ]["columns"][0]
                scaling_result_name = (self.scaling_ref, scaling_ref_column)
                if scaling_result_name not in result:
                    result_extract = result + (scaling_result_name,)
            data = self.get_data_for_result_tuple(
                n_matched_data, result_extract
            )
            yield data, result

    def get_data_for_result_tuple(self, n_matched_data, result_tuple):
        """
        Extract a dataframe for a given result tuple from the
        matched dataframes.

        Parameters
        ----------
        n_matched_data: dict of pandas.DataFrames
            DataFrames in which n datasets were temporally matched.
            The key is a tuple of the dataset names.
        result_tuple: tuple
            Tuple describing which datasets and columns should be
            extracted. ((dataset_name, column_name),
                        (dataset_name2, column_name2))

        Returns
        -------
        data: pd.DataFrame
            pandas DataFrame with columns extracted from the
            temporally matched datasets
        """
        # find the key into the temporally matched dataset by combining the
        # dataset parts of the result_names
        dskey = []
        for i, r in enumerate(result_tuple):
            dskey.append(r[0])

        dskey = tuple(dskey)
        if len(list(n_matched_data)[0]) == len(dskey):
            # we should have an exact match of datasets and
            # temporal matches

            # still need to make sure that dskey is in the right order and
            # contains all the same datasets as the n_matched_data
            for key in n_matched_data:
                if sorted(dskey) == sorted(key):
                    dskey = key
                    break
            if dskey not in n_matched_data:
                return []
            else:
                data = n_matched_data[dskey]
        else:
            # more datasets were temporally matched than are
            # requested now so we select a temporally matched
            # dataset that has the first key in common with the
            # temporal reference.

            # This guarantees that we only select columns from dataframes for
            # which the temporal reference dataset was included in the
            # temporal matching

            first_match = [
                key for key in n_matched_data if self.temporal_ref == key[0]
            ]
            found_key = None
            for key in first_match:
                for dsk in dskey:
                    if dsk not in key:
                        continue
                found_key = key
            data = n_matched_data[found_key]

        # extract only the relevant columns from matched DataFrame
        data = data[[x for x in result_tuple]]
        # drop values if one column is NaN
        data = data.dropna()
        return data

    def get_processing_jobs(self):
        """
        Returns processing jobs that this process can understand.

        Returns
        -------
        jobs : list
            List of cells or gpis to process.
        """
        jobs = []
        if self.data_manager.reference_grid is not None:
            if type(self.data_manager.reference_grid) is CellGrid:
                cells = self.data_manager.reference_grid.get_cells()
                for cell in cells:
                    (
                        cell_gpis,
                        cell_lons,
                        cell_lats,
                    ) = self.data_manager.reference_grid.grid_points_for_cell(
                        cell
                    )
                    jobs.append([cell_gpis, cell_lons, cell_lats])
            else:
                (
                    gpis,
                    lons,
                    lats,
                ) = self.data_manager.reference_grid.get_grid_points()
                jobs = [gpis, lons, lats]

        return jobs


def args_to_iterable(*args, **kwargs):
    """
    Convert arguments to iterables.


    Parameters
    ----------
    args: iterables or not
        arguments
    n : int, optional
        number of explicit arguments
    """
    if "n" in kwargs:
        n = kwargs["n"]
    else:
        n = len(args)

    arguments = []
    for i, arg in enumerate(args):
        arguments.append(ensure_iterable(arg))

    it = iter(arguments)
    for _ in range(n):
        yield next(it, None)
    if n < len(args):
        yield tuple(it)



# Copyright (c) 2014,Vienna University of Technology,
# Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of the Vienna University of Technology,
#      Department of Geodesy and Geoinformation nor the
#      names of its contributors may be used to endorse or promote products
#      derived from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''
Created on Mar 7, 2014

Plot anomalies around climatology using colors

@author: Christoph Paulik christoph.paulik@geo.tuwien.ac.at
'''

import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import pandas as pd

import pytesmo.time_series.anomaly as anom


def plot_clim_anom(df, clim=None, axes=None, markersize=0.75,
                   mfc='0.3', mec='0.3', clim_color='0.0',
                   clim_linewidth=0.5, clim_linestyle='-',
                   pos_anom_color='#799ADA', neg_anom_color='#FD8086',
                   anom_linewidth=0.2, add_titles=True):
    """
    Takes a pandas DataFrame and calculates the climatology and anomaly
    and plots them in a nice way for each column

    Parameters
    ----------
    df : pandas.DataFrame
    clim : pandas.DataFrame, optional
        if given these climatologies will be used
        if not given then climatologies will be calculated
        this DataFrame must have the same number of columns as df
        and also the column names.
        each climatology must have doy as index.
    axes : list of matplotlib.Axes, optional
           list of axes on which each column should be plotted
           if not given a standard layout is generated
    markersize : float, optional
        size of the markers for the datapoints
    mfc : matplotlib color, optional
        markerfacecolor, color of the marker face
    mec : matplotlib color, optional
        markeredgecolor
    clim_color : matplotlib color, optional
        color of the climatology
    clim_linewidth : float, optional
        linewidth of the climatology
    clim_linestyle : string, optional
        linestyle of the climatology
    pos_anom_color : matplotlib color, optional
        color of the positive anomaly
    neg_anom_color : matplotlib color, optional
        color of the negative anomaly
    anom_linewidth : float, optional
        linewidth of the anomaly lines
    add_titles : boolean, optional
        if set each subplot will have it's column name as title
        Default : True

    Returns
    -------
    Figure : matplotlib.Figure
        if no axes were given
    axes : list of matploblib.Axes
        if no axes were given
    """

    if type(df) == pd.Series:
        df = pd.DataFrame(df)

    nr_columns = len(df.columns)

    # make own axis if necessary
    if axes is None:
        own_axis = True
        gs = gridspec.GridSpec(nr_columns, 1, right=0.8)

        fig = plt.figure(num=None, figsize=(6, 2 * nr_columns),
                         dpi=150, facecolor='w', edgecolor='k')

        last_axis = fig.add_subplot(gs[nr_columns - 1])
        axes = []
        for i, grid in enumerate(gs):
            if i < nr_columns - 1:
                ax = fig.add_subplot(grid, sharex=last_axis)
                axes.append(ax)
                ax.xaxis.set_visible(False)
        axes.append(last_axis)

    else:
        own_axis = False

    for i, column in enumerate(df):
        Ser = df[column]
        ax = axes[i]

        if clim is None:
            clima = anom.calc_climatology(Ser)
        else:
            clima = pd.Series(clim[column])
        anomaly = anom.calc_anomaly(Ser, climatology=clima, return_clim=True)

        anomaly[Ser.name] = Ser
        anomaly = anomaly.dropna()

        pos_anom = anomaly[Ser.name].values > anomaly['climatology'].values
        neg_anom = anomaly[Ser.name].values < anomaly['climatology'].values

        ax.plot(anomaly.index, anomaly[Ser.name].values, 'o',
                markersize=markersize, mfc=mfc, mec=mec)

        ax.plot(anomaly.index, anomaly['climatology'].values,
                linestyle=clim_linestyle,
                color=clim_color,
                linewidth=clim_linewidth)

        ax.fill_between(anomaly.index,
                        anomaly[Ser.name].values,
                        anomaly['climatology'].values, interpolate=True,
                        where=pos_anom, color=pos_anom_color,
                        linewidth=anom_linewidth)
        ax.fill_between(anomaly.index,
                        anomaly[Ser.name].values,
                        anomaly['climatology'].values, interpolate=True,
                        where=neg_anom, color=neg_anom_color,
                        linewidth=anom_linewidth)
        if add_titles:
            ax.set_title(column)

    if own_axis:
        return fig, axes
    else:
        return None, None



# Copyright (c) 2015,Vienna University of Technology,
# Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of the Vienna University of Technology,
#      Department of Geodesy and Geoinformation nor the
#      names of its contributors may be used to endorse or promote products
#      derived from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''
Test for timeseries plotting
'''

import numpy as np
import pandas as pd
import matplotlib
matplotlib.use('Agg')
import pytest

import pytesmo.time_series.plotting as plotting


@pytest.mark.xfail(matplotlib.__version__ != '2.1.2', reason='plotting defaults change')
@pytest.mark.mpl_image_compare(tolerance=21)
def test_anomaly_calc_given_climatology():

    clim = pd.DataFrame({'data': np.concatenate((np.arange(150) - 40,
                                                 np.arange(216) + 190))},
                        index=np.arange(366) + 1)
    data = pd.DataFrame({'data': np.arange(366)},
                        index=pd.date_range('2000-01-01', periods=366))
    fig, axes = plotting.plot_clim_anom(data, clim=clim)
    return fig





import numpy as np
import pandas as pd
from scipy import interpolate, optimize, special
from sklearn.base import BaseEstimator, RegressorMixin
from typing import Sequence
import warnings

from typing import Union


class CDFMatching(RegressorMixin, BaseEstimator):
    """
    Predicts variable from single other variable by CDF matching.

    Parameters
    ----------
    nbins : int, optional
        Number of bins to use for the empirical CDF. Default is 100. If
        `minobs` is set, this might be reduced in case there's not enough data
        in each bin.
    percentiles : sequence, optional
        Percentile values to use. If this is given, `nbins` is ignored. The
        percentiles might still be changed if `minobs` is given and the number
        data per bin is lower. Default is ``None``.
    linear_edge_scaling : bool, optional
        Whether to derive the edge parameters via linear regression (more
        robust, see Moesinger et al. (2020) for more info). Default is
        ``False``.
        Note that this way only the outliers in the reference (y) CDF are
        handled. Outliers in the input data (x) will not be removed and will
        still show up in the data.
    minobs : int, optional
        Minimum desired number of observations in a bin. If there is less data
        for a bin, the number of bins is reduced. Default is ``None`` (no
        resizing).
    combine_invalid : bool, optional
        Optional feature to combine the masks of invalid data (NaN, Inf) of
        both source (X) and reference (y) data passed to `fit`. This only makes
        sense if X and y are both timeseries data corresponding to the same
        index. In this case, this makes sures that data is only used if values
        for X and y are available, so that seasonal patterns in missing values
        in one of them do not lead to distortions. (For example, if X is
        available the whole year, but y is only available during summer, the
        distribution of y should not be matched against the whole year CDF of
        X, because that could introduce systematic seasonal biases).
        The default is ``False``.

    Attributes
    ----------
    x_perc_ : np.ndarray (nbins,)
        The percentile values derived from the source (X) data. If the number
        of bins was reduced during fitting due to insufficient data, it is
        right-padded with NaNs.
    y_perc_ : np.ndarray (nbins,)
        The percentile values derived from the reference (y) data. If the
        number of bins was reduced during fitting due to insufficient data, it
        is right-padded with NaNs.

    Notes
    -----
    This implementation does not do any temporal matching of the reference and
    source datasets. If this is required, this has to be done beforehand.
    """

    def __init__(
        self,
        nbins: int = 100,
        percentiles: Sequence = None,
        minobs: int = None,
        linear_edge_scaling: bool = False,
        combine_invalid: bool = False,
    ):
        self.nbins = nbins
        self.minobs = minobs
        self.linear_edge_scaling = linear_edge_scaling
        self.combine_invalid = combine_invalid
        self.percentiles = percentiles
        if self.percentiles is not None:
            self.nbins = len(self.percentiles) - 1

    def fit(
        self,
        X: Union[np.ndarray, pd.Series, pd.DataFrame],
        y: Union[np.ndarray, pd.Series],
    ):
        """
        Derive the CDF matching parameters.

        Parameters
        ----------
        X : array_like
           An array/pd.Series or a matrix/pd.DataFrame with a single column.
        y : array_like
           An array/pd.Series of reference data.
        """

        # make sure that x and y are 1D numpy arrays
        if isinstance(y, pd.Series):
            y = y.values
        x = _make_X_array(X)

        # drop invalid values and pre-sort to avoid multiple resorting later
        isvalid_x = np.isfinite(x)
        isvalid_y = np.isfinite(y)
        if len(x) == len(y) and self.combine_invalid:
            isvalid_x = isvalid_x & isvalid_y
            isvalid_y = isvalid_x
        x_unsorted = x[isvalid_x]
        y_unsorted = y[isvalid_y]
        x = np.sort(x_unsorted)
        y = np.sort(y_unsorted)

        # calculate percentiles, potentially resize them
        nsamples = min(len(x), len(y))
        tmp_percentiles, tmp_nbins = self._calc_percentiles(nsamples)

        if tmp_nbins == 1 and self.linear_edge_scaling:
            # for a single bin just do a linear interpolation in case linear
            # edge scaling is activated
            x_perc, y_perc = _linreg_percentiles(x_unsorted, y_unsorted)
        else:
            x_perc = _percentile_values_from_sorted(x, tmp_percentiles)
            y_perc = _percentile_values_from_sorted(y, tmp_percentiles)
            if self.linear_edge_scaling:
                y_perc = self._linear_edge_scaling(
                    x, y, x_perc, y_perc, tmp_percentiles)

        # fill self.x_perc_ and self.y_perc_ with NaN on the right in case the
        # bin size was reduced, so we always get arrays of size nbins as
        # self.x_perc_ and self.y_perc_
        self.x_perc_ = np.full(self.nbins + 1, np.nan, dtype=x.dtype)
        self.y_perc_ = np.full(self.nbins + 1, np.nan, dtype=x.dtype)
        self.percentiles_ = np.full(self.nbins + 1, np.nan, dtype=x.dtype)
        self.percentiles_[0:tmp_nbins + 1] = tmp_percentiles
        self.x_perc_[0:tmp_nbins + 1] = x_perc
        self.y_perc_[0:tmp_nbins + 1] = y_perc
        return self

    def predict(self, X):
        x = _make_X_array(X)
        isvalid = np.isfinite(x)
        prediction = np.full_like(x, np.nan)
        if not np.any(isvalid):
            return prediction
        xp = self.x_perc_[~np.isnan(self.x_perc_)]
        yp = self.y_perc_[~np.isnan(self.y_perc_)]
        if len(xp) == 0 or len(yp) == 0:
            return prediction
        else:
            try:
                spline = interpolate.InterpolatedUnivariateSpline(
                    xp, yp, k=1, ext=0)
                prediction[isvalid] = spline(x[isvalid])
                return prediction
            except ValueError:
                # happens if there are non-unique values or not enough values
                warnings.warn("Too few percentiles for chosen k.")
                return prediction

    def _calc_percentiles(self, nsamples):
        # calculate percentiles, potentially resize them
        if self.percentiles is None:
            percentiles = np.arange(
                self.nbins + 1, dtype=np.float64) / self.nbins * 100
        else:
            percentiles = self.percentiles

        minbinsize = np.min(np.diff(percentiles))
        if (self.minobs is not None and
                nsamples * minbinsize / 100 < self.minobs):
            warnings.warn("The bins have been resized")
            nbins = np.int32(np.floor(nsamples / self.minobs))
            if nbins == 0:
                nbins = 1
            elif nbins > len(percentiles) - 1:
                nbins = len(percentiles) - 1

            percentiles = np.arange(nbins + 1, dtype=np.float64) / nbins * 100
        else:
            nbins = len(percentiles) - 1
        return percentiles, nbins

    def _linear_edge_scaling(self, x, y, x_perc, y_perc, percentiles):
        # scales the lower and upper edges of y_perc by replacing the values
        # with values inferred from a linear regression between the n
        # lowest/highest values for x and y. n is the number of data in the
        # lowest/highest y-bin, if there are more or less values in the
        # corresponding x-bin, the values will be resampled from the empirical
        # CDF. This can happen if the values are on only a few discrete levels.
        xlow = x[x <= x_perc[1]] - x_perc[1]
        ylow = y[y <= y_perc[1]] - y_perc[1]
        n = len(ylow)
        if len(xlow) != n:
            xlow = _resample_ecdf(xlow, n)
        a, _, _, _ = np.linalg.lstsq(xlow.reshape(-1, 1), ylow, rcond=None)
        y_perc[0] = y_perc[1] + a[0] * (x_perc[0] - x_perc[1])

        xhigh = x[x >= x_perc[-2]] - x_perc[-2]
        yhigh = y[y >= y_perc[-2]] - y_perc[-2]
        n = len(yhigh)
        if len(xhigh) != n:
            xhigh = _resample_ecdf(xhigh, n)
        a, _, _, _ = np.linalg.lstsq(xhigh.reshape(-1, 1), yhigh, rcond=None)
        y_perc[-1] = y_perc[-2] + a[0] * (x_perc[-1] - x_perc[-2])

        return y_perc


def _make_X_array(X):
    if isinstance(X, (pd.Series, pd.DataFrame)):
        X = X.values
    if len(X.shape) > 1:
        assert len(X.shape) == 2
        assert X.shape[1] == 1
        x = X.ravel()
    else:
        x = X
    return x


def _resample_ecdf(x_sorted, n, is_sorted=True):
    """Resample ECDF to n bins"""
    # calculate percentiles for x_sorted
    if not is_sorted:
        x_sorted = np.sort(x_sorted)
    new_percentiles = np.arange(n, dtype=float) / (n - 1) * 100.0
    return _percentile_values_from_sorted(x_sorted, new_percentiles)


def _percentile_values_from_sorted(data, percentiles):
    perc = _matlab_percentile_values_from_sorted(data, percentiles)
    return _unique_percentile_interpolation(perc, percentiles)


def _matlab_percentile_values_from_sorted(data, percentiles):
    """
    Calculate percentiles in the way Matlab and IDL do it.

    By using interpolation between the lowest an highest rank and the
    minimum and maximum outside.

    Parameters
    ----------
    data: numpy.ndarray
        sorted input data
    percentiles: numpy.ndarray
        percentiles at which to calculate the values
    Returns
    -------
    perc: numpy.ndarray
        values of the percentiles
    """
    p_rank = 100.0 * (np.arange(data.size) + 0.5) / data.size
    perc = np.interp(percentiles, p_rank, data, left=data[0], right=data[-1])
    return perc


def _unique_percentile_interpolation(perc, percentiles):
    # make sure the percentiles are unique
    # assumes `perc` is sorted
    uniq, uniq_ind = np.unique(perc, return_index=True)
    if uniq_ind.size != len(percentiles) and uniq_ind.size > 1:
        # If there are non-unique percentile values in perc, e.g.
        # [1, 1, 1, 2, ...] corresponding to percentiles [0, 5, 10, 20, ...],
        # we will interpolate the values for 5 and 10 to be between 1 and 2.
        # uniq_ind contains the first index of the non-unique values, so
        # selecting them will return [1, 2] and [0, 20].
        new_percentiles = percentiles[uniq_ind]
        new_perc = perc[uniq_ind]
        # However, if we have non-unique indices at the end of the array, e.g.
        # [..., 8, 10, 10, 10] corresponding to percentiles [..., 85, 90, 95,
        # 100], this approach will return [8, 10] and [85, 90], meaning that
        # the last percentile values will be extrapolated.
        # To avoid this, we set the last non-unique percentile to the overall
        # last percentile
        new_percentiles[-1] = percentiles[-1]
        spline = interpolate.InterpolatedUnivariateSpline(
            new_percentiles, new_perc, k=1, ext=0)
        perc = spline(percentiles)
    return perc


def _linreg_percentiles(x, y):
    """Calculate percentiles via linear regression."""
    A = np.reshape(x - np.mean(x), (-1, 1))
    b = y - np.mean(y)
    slope, _, _, _ = np.linalg.lstsq(A, b, rcond=None)
    intercept = np.mean(y) - np.mean(x) * slope[0]
    x_perc = [0, 1]
    y_perc = [intercept, intercept + slope[0]]
    return x_perc, y_perc



# Copyright (c) 2013,Vienna University of Technology, Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of the Vienna University of Technology, Department of Geodesy and Geoinformation nor the
#      names of its contributors may be used to endorse or promote products
#      derived from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''
Created on Oct 16, 2013

@author: Christoph Paulik christoph.paulik@geo.tuwien.ac.at
'''

import pandas as pd
import numpy as np

from cadati.jd_date import julday
from pytesmo.time_series.filters import boxcar_filter


def moving_average(Ser,
                   window_size=1,
                   fillna=False,
                   min_obs=1):
    '''
    Applies a moving average (box) filter on an input time series

    Parameters
    ----------
    Ser : pandas.Series (index must be a DateTimeIndex or julian date)

    window_size : float, optional
        The size of the moving_average window [days] that will be applied on the
        input Series
        Default: 1
    fillna: bool, optional
        Fill nan values at the center window value
    min_obs: int
        The minimum amount of observations necessary for a valid moving average

    Returns
    -------
    Ser : pandas.Series
        moving-average filtered time series
    '''
    # if index is datetimeindex then convert it to julian date
    if type(Ser.index) == pd.DatetimeIndex:

        jd_index = julday(np.asarray(Ser.index.month),
                          np.asarray(Ser.index.day),
                          np.asarray(Ser.index.year),
                          np.asarray(Ser.index.hour),
                          np.asarray(Ser.index.minute),
                          np.asarray(Ser.index.second))

    else:
        jd_index = Ser.index.values

    filtered = boxcar_filter(
        np.atleast_1d(np.squeeze(Ser.values.astype(np.double))),
        jd_index.astype(np.double),
        window=window_size, fillna=fillna, min_obs=min_obs)

    result = pd.Series(filtered, index=Ser.index)

    return result



# Copyright (c) 2020, TU Wien, Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of the TU Wien, Department of Geodesy and
#      Geoinformation nor the names of its contributors may be used to endorse
#      or promote products derived from this software without specific prior
#      written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Data scaler classes to be used together with the validation framework.
"""

import numpy as np
import pandas as pd
import pytesmo.scaling as scaling
from pytesmo.cdf_matching import CDFMatching
from pynetcf.point_data import GriddedPointData


class DefaultScaler(object):
    """
    Scaling class that implements the scaling based on a
    given method from the pytesmo.scaling module.

    Parameters
    ----------
    method: string
        The data will be scaled into the reference space using the
        method specified by this string.
    """

    def __init__(self, method):
        self.method = method

    def scale(self, data, reference_index, gpi_info):
        """
        Scale all columns in data to the
        column at the reference_index.

        Parameters
        ----------
        data: pandas.DataFrame
            temporally matched dataset
        reference_index: int
            Which column of the data contains the
            scaling reference.
        gpi_info: tuple
            tuple of at least, (gpi, lon, lat)
            Where gpi has to be the grid point indices
            of the grid of this scaler.

        Raises
        ------
        ValueError
            if scaling is not successful
        """
        return scaling.scale(
            data, method=self.method, reference_index=reference_index
        )


class CDFStoreParamsScaler(object):
    """
    CDF scaling using stored parameters if available.
    If stored parameters are not available they are calculated
    and written to disk.

    Parameters
    ----------
    path : string
        Path where the data is/should be stored
    grid : :py:class:`pygeogrids.grids.CellGrid` instance
        Grid on which the data is stored.
        Should be the same as the spatial reference grid
        of the validation framework instance in which this
        scaler is used.
    percentiles : list or np.ndarray
        Percentiles to use for CDF matching
    **matcher_kwargs : keyword arguments
        Passed on to :py:class:`pytesmo.cdf_matching.CDFMatching``
    """

    def __init__(
        self, path, grid, percentiles=[0, 5, 10, 30, 50, 70, 90, 95, 100],
        **matcher_kwargs
    ):
        self.path = path
        self.grid = grid
        self.percentiles = np.asanyarray(percentiles)
        self.io = GriddedPointData(
            path,
            grid,
            mode="a",
            ioclass_kws={
                "add_dims": {"src_ref": 2, "percentiles": len(percentiles)}
            },
        )
        self.matcher_kwargs = matcher_kwargs

    def scale(self, data, reference_index, gpi_info):
        """
        Scale all columns in data to the
        column at the reference_index.

        Parameters
        ----------
        data: pandas.DataFrame
            temporally matched dataset
        reference_index: int
            Which column of the data contains the
            scaling reference.
        gpi_info: tuple
            tuple of at least, (gpi, lon, lat)
            Where gpi has to be the grid point indices
            of the grid of this scaler.

        Raises
        ------
        ValueError
            if scaling is not successful
        """
        gpi = gpi_info[0]
        parameters = self.get_parameters(data, reference_index, gpi)

        refname = data.columns.values[reference_index]
        reference = data[refname]
        for column in data:
            if column == refname:
                continue
            params = parameters[f"{column}_{refname}"]
            matcher = CDFMatching(percentiles=self.percentiles,
                                  **self.matcher_kwargs)
            matcher.x_perc_ = params[0, :]
            matcher.y_perc_ = params[1, :]
            data[column] = pd.Series(matcher.predict(data[column]),
                                     index=data.index)

        return data

    def calc_parameters(self, data, reference_index):
        """
        Calculate the percentiles used for CDF matching.

        Parameters
        ----------
        data: pandas.DataFrame
            temporally matched dataset
        reference_index : int
            Index of the reference column in the dataset.

        Returns
        -------
        matchers: dictionary
            keys -> Names of columns in the input data frame
            values -> nbins x 3 numpy.ndarrays with columns x_perc, y_perc,
                      percentiles
        """

        parameters = {}
        refname = data.columns[reference_index]
        for column in data.columns:
            if column == refname:
                continue
            matcher = CDFMatching(percentiles=self.percentiles,
                                  **self.matcher_kwargs)
            matcher.fit(data[column], data[refname])
            nperc = matcher.nbins + 1
            params = np.zeros((2, nperc), matcher.x_perc_.dtype)
            params[0, :] = matcher.x_perc_
            params[1, :] = matcher.y_perc_
            parameters[f"{column}_{refname}"] = params
        return parameters

    def get_parameters(self, data, reference_index, gpi):
        """
        Function to get scaling parameters.
        Try to load them, if they are not found we
        calculate them and store them.

        Parameters
        ----------
        data: pandas.DataFrame
            temporally matched dataset
        gpi: int
            grid point index of self.grid

        Returns
        -------
        params: dictionary
            keys -> Names of columns in the input data frame
            values -> numpy.ndarrays with the percentiles
        """

        params = self.load_parameters(gpi)
        if params is None:
            params = self.calc_parameters(data, reference_index)
            self.store_parameters(gpi, params)
        return params

    def load_parameters(self, gpi):
        data = self.io.read(gpi)
        if data is not None:
            unwanted_keys = ["lat", "lon", "alt", "time", "location_id"]
            for key in unwanted_keys:
                del data[key]

            # remove extra dimension from reading from netCDF
            for key in data:
                data[key] = np.squeeze(data[key])

        return data

    def store_parameters(self, gpi, parameters):
        """
        Store parameters for gpi into netCDF file.

        Parameters
        ----------
        gpi: int
            grid point index of self.grid
        params: dictionary
            keys -> Names of columns in the input data frame
            values -> numpy.ndarrays with the percentiles
        """
        data = []
        dtypes = []
        dim_info = {"dims": {}}
        for key in parameters:
            dim_info["dims"][key] = ("obs", "src_ref", "percentiles")
            dtypes.append(
                (key, parameters[key].dtype, (parameters[key].shape[-1],))
            )
            data.append(parameters[key])

        data = np.core.records.fromarrays(
            data, dtype=np.dtype(dtypes, metadata=dim_info)
        )
        self.io.write(gpi, data)
